@@ -90,7 +90,7 @@ struct Switch {
 	int linearFDBtop;
 	int fdb_base;
 	int enhsp0;
-	int8_t fdb[64];
+	uint8_t fdb[64];
 	char switchinfo[64];
 };
 
@@ -105,7 +105,7 @@ struct Node {
 	int upport;
 	Node *upnode;
 	uint64_t nodeguid;	/* also portguid */
-	char nodedesc[64];
+	char nodedesc[IB_SMP_DATA_SIZE + 1];
 	char nodeinfo[64];
 };
 
@@ -542,11 +542,15 @@ static Node *find_mcpath(ib_portid_t * from, int mlid)
 
 	DEBUG("from %s", portid2str(from));
 
-	if (!(node = calloc(1, sizeof(Node))))
+	node = calloc(1, sizeof(Node));
+	if (!node)
 		IBEXIT("out of memory");
 
-	if (!(port = calloc(1, sizeof(Port))))
+	port = calloc(1, sizeof(Port));
+	if (!port) {
+		free(node);
 		IBEXIT("out of memory");
+	}
 
 	if (get_node(node, port, from) < 0) {
 		IBWARN("can't reach node %s", portid2str(from));
@@ -564,8 +568,11 @@ static Node *find_mcpath(ib_portid_t * from, int mlid)
 			return NULL;	/* ibtracert from host to itself is unsupported */
 		}
 
-		if (switch_mclookup(node, from, mlid, map) < 0 || !map[0])
+		if (switch_mclookup(node, from, mlid, map) < 0 || !map[0]) {
+			free(node);
+			free(port);
 			return NULL;
+		}
 		return node;
 	}
 
@@ -612,7 +619,8 @@ static Node *find_mcpath(ib_portid_t * from, int mlid)
 					if (from->drpath.cnt > 0)
 						path->drpath.cnt--;
 				} else {
-					if (!(port = calloc(1, sizeof(Port))))
+					port = calloc(1, sizeof(Port));
+					if (!port)
 						IBEXIT("out of memory");
 
 					if (get_port(port, i, path) < 0) {
@@ -637,11 +645,18 @@ static Node *find_mcpath(ib_portid_t * from, int mlid)
 					}
 				}
 
-				if (!(remotenode = calloc(1, sizeof(Node))))
+				remotenode = calloc(1, sizeof(Node));
+				if (!remotenode) {
+					free(port);
 					IBEXIT("out of memory");
+				}
 
-				if (!(remoteport = calloc(1, sizeof(Port))))
+				remoteport = calloc(1, sizeof(Port));
+				if (!remoteport) {
+					free(port);
+					free(remotenode);
 					IBEXIT("out of memory");
+				}
 
 				if (get_node(remotenode, remoteport, path) < 0) {
 					IBWARN
@@ -670,6 +685,9 @@ static Node *find_mcpath(ib_portid_t * from, int mlid)
 						     remotenode, remoteport);
 
 				path->drpath.cnt--;	/* restore path */
+				free(port);
+				free(remotenode);
+				free(remoteport);
 			}
 		}
 	}
