@@ -47,6 +47,7 @@
 #include <sys/param.h>
 
 #include <util/mmio.h>
+#include <util/util.h>
 #include <infiniband/driver.h>
 #include <util/udma_barrier.h>
 
@@ -61,12 +62,19 @@
 #define CHIP_NUM_57508		0x1750
 #define CHIP_NUM_57504		0x1751
 #define CHIP_NUM_57502		0x1752
+#define CHIP_NUM_58818          0xd818
+#define CHIP_NUM_57608          0x1760
+
+#define BNXT_RE_MAX_DO_PACING	0xFFFF
+#define BNXT_NSEC_PER_SEC	1000000000UL
+#define BNXT_RE_PAGE_MASK(pg_size) (~((__u64)(pg_size) - 1))
 
 struct bnxt_re_chip_ctx {
 	__u16 chip_num;
 	__u8 chip_rev;
 	__u8 chip_metal;
-	__u8 gen_p5;
+	__u8 gen_p5_p7;
+	__u8 gen_p7;
 };
 
 struct bnxt_re_dpi {
@@ -84,6 +92,7 @@ struct bnxt_re_pd {
 struct bnxt_re_cq {
 	struct ibv_cq ibvcq;
 	uint32_t cqid;
+	struct bnxt_re_context *cntx;
 	struct bnxt_re_queue cqq;
 	struct bnxt_re_queue resize_cqq;
 	struct bnxt_re_dpi *udpi;
@@ -92,6 +101,10 @@ struct bnxt_re_cq {
 	struct list_head prev_cq_head;
 	uint32_t cqe_size;
 	uint8_t  phase;
+	struct xorshift32_state rand;
+	uint32_t mem_handle;
+	void *toggle_map;
+	uint32_t toggle_size;
 };
 
 struct bnxt_re_push_buffer {
@@ -140,14 +153,17 @@ struct bnxt_re_qpcap {
 	uint32_t max_rsge;
 	uint32_t max_inline;
 	uint8_t	sqsig;
+	uint8_t is_atomic_cap;
 };
 
 struct bnxt_re_srq {
 	struct ibv_srq ibvsrq;
 	struct ibv_srq_attr cap;
+	struct bnxt_re_context *cntx;
 	struct bnxt_re_queue *srqq;
 	struct bnxt_re_wrid *srwrid;
 	struct bnxt_re_dpi *udpi;
+	struct xorshift32_state rand;
 	uint32_t srqid;
 	int start_idx;
 	int last_idx;
@@ -165,6 +181,7 @@ struct bnxt_re_qp {
 	struct ibv_qp ibvqp;
 	struct bnxt_re_chip_ctx *cctx;
 	struct bnxt_re_context *cntx;
+	struct xorshift32_state rand;
 	struct bnxt_re_joint_queue *jsqq;
 	struct bnxt_re_joint_queue *jrqq;
 	struct bnxt_re_srq *srq;
@@ -219,9 +236,20 @@ struct bnxt_re_context {
 	void *shpg;
 	uint32_t wqe_mode;
 	pthread_mutex_t shlock;
-	pthread_spinlock_t fqlock;
 	struct bnxt_re_push_rec *pbrec;
 	uint32_t wc_handle;
+	void *dbr_page;
+	void *bar_map;
+};
+
+struct bnxt_re_pacing_data {
+	uint32_t do_pacing;
+	uint32_t pacing_th;
+	uint32_t alarm_th;
+	uint32_t fifo_max_depth;
+	uint32_t fifo_room_mask;
+	uint32_t fifo_room_shift;
+	uint32_t grc_reg_offset;
 };
 
 struct bnxt_re_mmap_info {
@@ -229,6 +257,8 @@ struct bnxt_re_mmap_info {
 	__u32 dpi;
 	__u64 alloc_offset;
 	__u32 alloc_size;
+	__u32 pg_offset;
+	__u32 res_id;
 };
 
 /* DB ring functions used internally*/
@@ -253,6 +283,14 @@ struct bnxt_re_push_buffer *bnxt_re_get_pbuf(uint8_t *push_st_en,
 					     struct bnxt_re_context *cntx);
 void bnxt_re_put_pbuf(struct bnxt_re_context *cntx,
 		      struct bnxt_re_push_buffer *pbuf);
+int bnxt_re_alloc_page(struct ibv_context *ibvctx,
+		       struct bnxt_re_mmap_info *minfo,
+		       uint32_t *page_handle);
+int bnxt_re_notify_drv(struct ibv_context *ibvctx);
+int bnxt_re_get_toggle_mem(struct ibv_context *ibvctx,
+			   struct bnxt_re_mmap_info *minfo,
+			   uint32_t *page_handle);
+
 /* pointer conversion functions*/
 static inline struct bnxt_re_dev *to_bnxt_re_dev(struct ibv_device *ibvdev)
 {
@@ -513,6 +551,12 @@ static inline void bnxt_re_jqq_mod_last(struct bnxt_re_joint_queue *jqq,
 	jqq->last_idx = jqq->swque[idx].next_idx;
 }
 
+static inline uint32_t bnxt_re_init_depth(uint32_t ent, uint64_t cmask)
+{
+	return cmask & BNXT_RE_COMP_MASK_UCNTX_POW2_DISABLED ?
+		ent : roundup_pow_of_two(ent);
+}
+
 /* Helper function to copy to push buffers */
 static inline void bnxt_re_copy_data_to_pb(struct bnxt_re_push_buffer *pbuf,
 					   uint8_t offset, uint32_t idx)
@@ -531,4 +575,40 @@ static inline void bnxt_re_copy_data_to_pb(struct bnxt_re_push_buffer *pbuf,
 		mmio_write64(dst, *src);
 	}
 }
+
+static void timespec_sub(const struct timespec *a, const struct timespec *b,
+			 struct timespec *res)
+{
+	res->tv_sec = a->tv_sec - b->tv_sec;
+	res->tv_nsec = a->tv_nsec - b->tv_nsec;
+	if (res->tv_nsec < 0) {
+		res->tv_sec--;
+		res->tv_nsec += BNXT_NSEC_PER_SEC;
+	}
+}
+
+/*
+ * Function waits in a busy loop for a given nano seconds
+ * The maximum wait period allowed is less than one second
+ */
+static inline void bnxt_re_sub_sec_busy_wait(uint32_t nsec)
+{
+	struct timespec start, cur, res;
+
+	if (nsec >= BNXT_NSEC_PER_SEC)
+		return;
+
+	if (clock_gettime(CLOCK_REALTIME, &start))
+		return;
+
+	while (1) {
+		if (clock_gettime(CLOCK_REALTIME, &cur))
+			return;
+		timespec_sub(&cur, &start, &res);
+		if (res.tv_nsec >= nsec)
+			break;
+	}
+}
+
+#define BNXT_RE_HW_RETX(a) ((a)->comp_mask & BNXT_RE_COMP_MASK_UCNTX_HW_RETX_ENABLED)
 #endif
