@@ -44,12 +44,15 @@
 
 struct bnxt_re_queue {
 	void *va;
+	uint32_t flags;
 	uint32_t *dbtail;
 	uint32_t bytes; /* for munmap */
 	uint32_t depth; /* no. of entries */
 	uint32_t head;
 	uint32_t tail;
 	uint32_t stride;
+	void *pad; /* to hold the padding area */
+	uint32_t pad_stride_log2;
 	/* Represents the difference between the real queue depth allocated in
 	 * HW and the user requested queue depth and is used to correctly flag
 	 * queue full condition based on user supplied queue depth.
@@ -61,26 +64,10 @@ struct bnxt_re_queue {
 	uint32_t esize;
 	uint32_t max_slots;
 	pthread_spinlock_t qlock;
+	uint32_t msn;
+	uint32_t msn_tbl_sz;
 };
 
-static inline unsigned long get_aligned(uint32_t size, uint32_t al_size)
-{
-	return (unsigned long)(size + al_size - 1) & ~(al_size - 1);
-}
-
-static inline unsigned long roundup_pow_of_two(unsigned long val)
-{
-	unsigned long roundup = 1;
-
-	if (val == 1)
-		return (roundup << 1);
-
-	while (roundup < val)
-		roundup <<= 1;
-
-	return roundup;
-}
-
 int bnxt_re_alloc_aligned(struct bnxt_re_queue *que, uint32_t pg_size);
 void bnxt_re_free_aligned(struct bnxt_re_queue *que);
 
@@ -119,15 +106,21 @@ static inline uint32_t bnxt_re_is_que_empty(struct bnxt_re_queue *que)
 static inline void bnxt_re_incr_tail(struct bnxt_re_queue *que, uint8_t cnt)
 {
 	que->tail += cnt;
-	if (que->tail >= que->depth)
+	if (que->tail >= que->depth) {
 		que->tail %= que->depth;
+		/* Rolled over, Toggle Tail bit in epoch flags */
+		que->flags ^= 1UL << BNXT_RE_FLAG_EPOCH_TAIL_SHIFT;
+	}
 }
 
 static inline void bnxt_re_incr_head(struct bnxt_re_queue *que, uint8_t cnt)
 {
 	que->head += cnt;
-	if (que->head >= que->depth)
+	if (que->head >= que->depth) {
 		que->head %= que->depth;
+		/* Rolled over, Toggle HEAD bit in epoch flags */
+		que->flags ^= 1UL << BNXT_RE_FLAG_EPOCH_HEAD_SHIFT;
+	}
 }
 
 #endif
