@@ -47,8 +47,10 @@
 #include <sys/mman.h>
 #include <netinet/in.h>
 #include <unistd.h>
+#include <ccan/ilog.h>
 
 #include <util/compiler.h>
+#include <util/util.h>
 
 #include "main.h"
 #include "verbs.h"
@@ -102,9 +104,48 @@ static int bnxt_re_map_db_page(struct ibv_context *ibvctx,
 	return 0;
 }
 
-static int bnxt_re_alloc_page(struct ibv_context *ibvctx,
-			      struct bnxt_re_mmap_info *minfo,
-			      uint32_t *page_handle)
+int bnxt_re_get_toggle_mem(struct ibv_context *ibvctx,
+			   struct bnxt_re_mmap_info *minfo,
+			   uint32_t *page_handle)
+{
+	DECLARE_COMMAND_BUFFER(cmd,
+			       BNXT_RE_OBJECT_GET_TOGGLE_MEM,
+			       BNXT_RE_METHOD_GET_TOGGLE_MEM,
+			       4);
+	struct ib_uverbs_attr *handle;
+	int ret;
+
+	handle = fill_attr_out_obj(cmd, BNXT_RE_TOGGLE_MEM_HANDLE);
+	fill_attr_const_in(cmd, BNXT_RE_TOGGLE_MEM_TYPE, minfo->type);
+	fill_attr_in(cmd, BNXT_RE_TOGGLE_MEM_RES_ID, &minfo->res_id, sizeof(minfo->res_id));
+	fill_attr_out_ptr(cmd, BNXT_RE_TOGGLE_MEM_MMAP_PAGE,  &minfo->alloc_offset);
+	fill_attr_out_ptr(cmd, BNXT_RE_TOGGLE_MEM_MMAP_LENGTH, &minfo->alloc_size);
+	fill_attr_out_ptr(cmd, BNXT_RE_TOGGLE_MEM_MMAP_OFFSET, &minfo->pg_offset);
+
+
+	ret = execute_ioctl(ibvctx, cmd);
+
+	if (ret)
+		return ret;
+	if (page_handle)
+		*page_handle = read_attr_obj(BNXT_RE_TOGGLE_MEM_HANDLE, handle);
+	return 0;
+}
+
+
+int bnxt_re_notify_drv(struct ibv_context *ibvctx)
+{
+	DECLARE_COMMAND_BUFFER(cmd,
+			       BNXT_RE_OBJECT_NOTIFY_DRV,
+			       BNXT_RE_METHOD_NOTIFY_DRV,
+			       0);
+
+	return execute_ioctl(ibvctx, cmd);
+}
+
+int bnxt_re_alloc_page(struct ibv_context *ibvctx,
+		       struct bnxt_re_mmap_info *minfo,
+		       uint32_t *page_handle)
 {
 	DECLARE_COMMAND_BUFFER(cmd,
 			       BNXT_RE_OBJECT_ALLOC_PAGE,
@@ -124,7 +165,8 @@ static int bnxt_re_alloc_page(struct ibv_context *ibvctx,
 
 	if (ret)
 		return ret;
-	*page_handle = read_attr_obj(BNXT_RE_ALLOC_PAGE_HANDLE, handle);
+	if (page_handle)
+		*page_handle = read_attr_obj(BNXT_RE_ALLOC_PAGE_HANDLE, handle);
 	return 0;
 }
 
@@ -178,7 +220,7 @@ struct ibv_pd *bnxt_re_alloc_pd(struct ibv_context *ibvctx)
 			goto fail;
 		if (bnxt_re_is_wcdpi_enabled(cntx)) {
 			bnxt_re_alloc_map_push_page(ibvctx);
-			if (cntx->cctx.gen_p5 && cntx->udpi.wcdpi)
+			if (cntx->cctx.gen_p5_p7 && cntx->udpi.wcdpi)
 				bnxt_re_init_pbuf_list(cntx);
 		}
         }
@@ -225,6 +267,24 @@ struct ibv_mr *bnxt_re_reg_mr(struct ibv_pd *ibvpd, void *sva, size_t len,
 	return &mr->vmr.ibv_mr;
 }
 
+struct ibv_mr *bnxt_re_reg_dmabuf_mr(struct ibv_pd *ibvpd, uint64_t start, size_t len,
+				     uint64_t iova, int fd, int access)
+{
+	struct bnxt_re_mr *mr;
+
+	mr = calloc(1, sizeof(*mr));
+	if (!mr)
+		return NULL;
+
+	if (ibv_cmd_reg_dmabuf_mr(ibvpd, start, len, iova, fd,
+				  access, &mr->vmr)) {
+		free(mr);
+		return NULL;
+	}
+
+	return &mr->vmr.ibv_mr;
+}
+
 int bnxt_re_dereg_mr(struct verbs_mr *vmr)
 {
 	struct bnxt_re_mr *mr = (struct bnxt_re_mr *)vmr;
@@ -244,6 +304,8 @@ struct ibv_cq *bnxt_re_create_cq(struct ibv_context *ibvctx, int ncqe,
 	struct bnxt_re_cq *cq;
 	struct ubnxt_re_cq cmd;
 	struct ubnxt_re_cq_resp resp;
+	struct bnxt_re_mmap_info minfo = {};
+	int ret;
 
 	struct bnxt_re_context *cntx = to_bnxt_re_context(ibvctx);
 	struct bnxt_re_dev *dev = to_bnxt_re_dev(ibvctx->device);
@@ -257,7 +319,7 @@ struct ibv_cq *bnxt_re_create_cq(struct ibv_context *ibvctx, int ncqe,
 	if (!cq)
 		return NULL;
 
-	cq->cqq.depth = roundup_pow_of_two(ncqe + 1);
+	cq->cqq.depth = bnxt_re_init_depth(ncqe + 1, cntx->comp_mask);
 	if (cq->cqq.depth > dev->max_cq_depth + 1)
 		cq->cqq.depth = dev->max_cq_depth + 1;
 	cq->cqq.stride = dev->cqe_size;
@@ -279,7 +341,22 @@ struct ibv_cq *bnxt_re_create_cq(struct ibv_context *ibvctx, int ncqe,
 	cq->phase = resp.phase;
 	cq->cqq.tail = resp.tail;
 	cq->udpi = &cntx->udpi;
+	cq->cntx = cntx;
+	cq->rand.seed = cq->cqid;
 
+	if (resp.comp_mask & BNXT_RE_CQ_TOGGLE_PAGE_SUPPORT) {
+
+		minfo.type = BNXT_RE_CQ_TOGGLE_MEM;
+		minfo.res_id = resp.cqid;
+		ret = bnxt_re_get_toggle_mem(ibvctx, &minfo, &cq->mem_handle);
+		if (ret)
+			goto cmdfail;
+		cq->toggle_map = mmap(NULL, minfo.alloc_size, PROT_READ,
+				MAP_SHARED, ibvctx->cmd_fd, minfo.alloc_offset);
+		if (cq->toggle_map == MAP_FAILED)
+			goto cmdfail;
+		cq->toggle_size = minfo.alloc_size;
+	}
 	list_head_init(&cq->sfhead);
 	list_head_init(&cq->rfhead);
 	list_head_init(&cq->prev_cq_head);
@@ -311,6 +388,7 @@ static void bnxt_re_resize_cq_complete(struct bnxt_re_cq *cq)
 
 int bnxt_re_resize_cq(struct ibv_cq *ibvcq, int ncqe)
 {
+	struct bnxt_re_context *cntx = to_bnxt_re_context(ibvcq->context);
 	struct bnxt_re_dev *dev = to_bnxt_re_dev(ibvcq->context->device);
 	struct bnxt_re_cq *cq = to_bnxt_re_cq(ibvcq);
 	struct ib_uverbs_resize_cq_resp resp = {};
@@ -321,7 +399,7 @@ int bnxt_re_resize_cq(struct ibv_cq *ibvcq, int ncqe)
 		return -EINVAL;
 
 	pthread_spin_lock(&cq->cqq.qlock);
-	cq->resize_cqq.depth = roundup_pow_of_two(ncqe + 1);
+	cq->resize_cqq.depth = bnxt_re_init_depth(ncqe + 1, cntx->comp_mask);
 	if (cq->resize_cqq.depth > dev->max_cq_depth + 1)
 		cq->resize_cqq.depth = dev->max_cq_depth + 1;
 	cq->resize_cqq.stride = dev->cqe_size;
@@ -380,6 +458,8 @@ int bnxt_re_destroy_cq(struct ibv_cq *ibvcq)
 	int status;
 	struct bnxt_re_cq *cq = to_bnxt_re_cq(ibvcq);
 
+	if (cq->toggle_map)
+		munmap(cq->toggle_map, cq->toggle_size);
 	status = ibv_cmd_destroy_cq(ibvcq);
 	if (status)
 		return status;
@@ -396,9 +476,7 @@ static uint8_t bnxt_re_poll_err_scqe(struct bnxt_re_qp *qp,
 				     struct bnxt_re_req_cqe *scqe, int *cnt)
 {
 	struct bnxt_re_queue *sq = qp->jsqq->hwque;
-	struct bnxt_re_context *cntx;
 	struct bnxt_re_wrid *swrid;
-	struct bnxt_re_psns *spsn;
 	struct bnxt_re_cq *scq;
 	uint8_t status;
 	uint32_t head;
@@ -406,9 +484,7 @@ static uint8_t bnxt_re_poll_err_scqe(struct bnxt_re_qp *qp,
 	scq = to_bnxt_re_cq(qp->ibvqp.send_cq);
 
 	head = qp->jsqq->last_idx;
-	cntx = to_bnxt_re_context(scq->ibvcq.context);
 	swrid = &qp->jsqq->swque[head];
-	spsn = swrid->psns;
 
 	*cnt = 1;
 	status = (le32toh(hdr->flg_st_typ_ph) >> BNXT_RE_BCQE_STATUS_SHIFT) &
@@ -417,9 +493,7 @@ static uint8_t bnxt_re_poll_err_scqe(struct bnxt_re_qp *qp,
 	ibvwc->wc_flags = 0;
 	ibvwc->wr_id = swrid->wrid;
 	ibvwc->qp_num = qp->qpid;
-	ibvwc->opcode = (le32toh(spsn->opc_spsn) >>
-			BNXT_RE_PSNS_OPCD_SHIFT) &
-			BNXT_RE_PSNS_OPCD_MASK;
+	ibvwc->opcode = swrid->wc_opcd;
 	ibvwc->byte_len = 0;
 
 	bnxt_re_incr_head(sq, swrid->slots);
@@ -427,9 +501,7 @@ static uint8_t bnxt_re_poll_err_scqe(struct bnxt_re_qp *qp,
 
 	if (qp->qpst != IBV_QPS_ERR)
 		qp->qpst = IBV_QPS_ERR;
-	pthread_spin_lock(&cntx->fqlock);
 	bnxt_re_fque_add_node(&scq->sfhead, &qp->snode);
-	pthread_spin_unlock(&cntx->fqlock);
 
 	return false;
 }
@@ -442,15 +514,12 @@ static uint8_t bnxt_re_poll_success_scqe(struct bnxt_re_qp *qp,
 {
 	struct bnxt_re_queue *sq = qp->jsqq->hwque;
 	struct bnxt_re_wrid *swrid;
-	struct bnxt_re_psns *spsn;
-	uint8_t pcqe = false;
 	uint32_t cindx;
 	uint32_t head;
 
 	head = qp->jsqq->last_idx;
 	swrid = &qp->jsqq->swque[head];
-	spsn = swrid->psns;
-	cindx = le32toh(scqe->con_indx) & (qp->cap.max_swr - 1);
+	cindx = le32toh(scqe->con_indx) % qp->cap.max_swr;
 
 	if (!(swrid->sig & IBV_SEND_SIGNALED)) {
 		*cnt = 0;
@@ -459,9 +528,7 @@ static uint8_t bnxt_re_poll_success_scqe(struct bnxt_re_qp *qp,
 		ibvwc->wc_flags = 0;
 		ibvwc->qp_num = qp->qpid;
 		ibvwc->wr_id = swrid->wrid;
-		ibvwc->opcode = (le32toh(spsn->opc_spsn) >>
-				BNXT_RE_PSNS_OPCD_SHIFT) &
-				BNXT_RE_PSNS_OPCD_MASK;
+		ibvwc->opcode = swrid->wc_opcd;
 		if (ibvwc->opcode == IBV_WC_RDMA_READ ||
 		    ibvwc->opcode == IBV_WC_COMP_SWAP ||
 		    ibvwc->opcode == IBV_WC_FETCH_ADD)
@@ -474,29 +541,27 @@ static uint8_t bnxt_re_poll_success_scqe(struct bnxt_re_qp *qp,
 	bnxt_re_jqq_mod_last(qp->jsqq, head);
 
 	if (qp->jsqq->last_idx != cindx)
-		pcqe = true;
+		return true;
 
-	return pcqe;
+	return false;
 }
 
 static uint8_t bnxt_re_poll_scqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 				 void *cqe, int *cnt)
 {
-	struct bnxt_re_bcqe *hdr;
 	struct bnxt_re_req_cqe *scqe;
-	uint8_t status, pcqe = false;
+	struct bnxt_re_bcqe *hdr;
+	uint8_t status;
 
 	scqe = cqe;
 	hdr = cqe + sizeof(struct bnxt_re_req_cqe);
 
 	status = (le32toh(hdr->flg_st_typ_ph) >> BNXT_RE_BCQE_STATUS_SHIFT) &
 		  BNXT_RE_BCQE_STATUS_MASK;
-	if (status == BNXT_RE_REQ_ST_OK)
-		pcqe = bnxt_re_poll_success_scqe(qp, ibvwc, hdr, scqe, cnt);
+	if (likely(status == BNXT_RE_REQ_ST_OK))
+		return bnxt_re_poll_success_scqe(qp, ibvwc, hdr, scqe, cnt);
 	else
-		pcqe = bnxt_re_poll_err_scqe(qp, ibvwc, hdr, scqe, cnt);
-
-	return pcqe;
+		return bnxt_re_poll_err_scqe(qp, ibvwc, hdr, scqe, cnt);
 }
 
 static void bnxt_re_release_srqe(struct bnxt_re_srq *srq, int tag)
@@ -511,7 +576,6 @@ static void bnxt_re_release_srqe(struct bnxt_re_srq *srq, int tag)
 static int bnxt_re_poll_err_rcqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 				 struct bnxt_re_bcqe *hdr, void *cqe)
 {
-	struct bnxt_re_context *cntx;
 	struct bnxt_re_wrid *swque;
 	struct bnxt_re_queue *rq;
 	uint8_t status, cnt = 0;
@@ -519,7 +583,6 @@ static int bnxt_re_poll_err_rcqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 	uint32_t head = 0;
 
 	rcq = to_bnxt_re_cq(qp->ibvqp.recv_cq);
-	cntx = to_bnxt_re_context(rcq->ibvcq.context);
 
 	if (!qp->srq) {
 		rq = qp->jrqq->hwque;
@@ -557,11 +620,8 @@ static int bnxt_re_poll_err_rcqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 		bnxt_re_jqq_mod_last(qp->jrqq, head);
 	bnxt_re_incr_head(rq, cnt);
 
-	if (!qp->srq) {
-		pthread_spin_lock(&cntx->fqlock);
+	if (!qp->srq)
 		bnxt_re_fque_add_node(&rcq->rfhead, &qp->rnode);
-		pthread_spin_unlock(&cntx->fqlock);
-	}
 
 	return 1;
 }
@@ -659,7 +719,7 @@ static uint8_t bnxt_re_poll_rcqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 	status = (le32toh(hdr->flg_st_typ_ph) >> BNXT_RE_BCQE_STATUS_SHIFT) &
 		  BNXT_RE_BCQE_STATUS_MASK;
 	*cnt = 1;
-	if (status == BNXT_RE_RSP_ST_OK)
+	if (likely(status == BNXT_RE_RSP_ST_OK))
 		bnxt_re_poll_success_rcqe(qp, ibvwc, hdr, cqe);
 	else
 		*cnt = bnxt_re_poll_err_rcqe(qp, ibvwc, hdr, cqe);
@@ -667,33 +727,31 @@ static uint8_t bnxt_re_poll_rcqe(struct bnxt_re_qp *qp, struct ibv_wc *ibvwc,
 	return pcqe;
 }
 
-static uint8_t bnxt_re_poll_term_cqe(struct bnxt_re_qp *qp,
-				     struct ibv_wc *ibvwc, void *cqe, int *cnt)
+static void bnxt_re_qp_move_flush_err(struct bnxt_re_qp *qp)
 {
-	struct bnxt_re_context *cntx;
 	struct bnxt_re_cq *scq, *rcq;
-	uint8_t pcqe = false;
 
 	scq = to_bnxt_re_cq(qp->ibvqp.send_cq);
 	rcq = to_bnxt_re_cq(qp->ibvqp.recv_cq);
-	cntx = to_bnxt_re_context(scq->ibvcq.context);
+
+	if (qp->qpst != IBV_QPS_ERR)
+		qp->qpst = IBV_QPS_ERR;
+	bnxt_re_fque_add_node(&rcq->rfhead, &qp->rnode);
+	bnxt_re_fque_add_node(&scq->sfhead, &qp->snode);
+}
+
+static uint8_t bnxt_re_poll_term_cqe(struct bnxt_re_qp *qp, int *cnt)
+{
 	/* For now just add the QP to flush list without
 	 * considering the index reported in the CQE.
 	 * Continue reporting flush completions until the
 	 * SQ and RQ are empty.
 	 */
 	*cnt = 0;
-	/* If the QP is destroyed, avoid handling this QP as flushlist */
-	if (qp->qpst == IBV_QPS_RESET)
-		goto exit;
-	if (qp->qpst != IBV_QPS_ERR)
-		qp->qpst = IBV_QPS_ERR;
-	pthread_spin_lock(&cntx->fqlock);
-	bnxt_re_fque_add_node(&rcq->rfhead, &qp->rnode);
-	bnxt_re_fque_add_node(&scq->sfhead, &qp->snode);
-	pthread_spin_unlock(&cntx->fqlock);
-exit:
-	return pcqe;
+	if (qp->qpst != IBV_QPS_RESET)
+		bnxt_re_qp_move_flush_err(qp);
+
+	return 0;
 }
 
 static int bnxt_re_poll_one(struct bnxt_re_cq *cq, int nwc, struct ibv_wc *wc,
@@ -745,7 +803,7 @@ static int bnxt_re_poll_one(struct bnxt_re_cq *cq, int nwc, struct ibv_wc *wc,
 			     (uintptr_t)le64toh(scqe->qp_handle);
 			if (!qp)
 				break;
-			pcqe = bnxt_re_poll_term_cqe(qp, wc, cqe, &cnt);
+			pcqe = bnxt_re_poll_term_cqe(qp, &cnt);
 			break;
 		case BNXT_RE_WC_TYPE_COFF:
 			/* Stop further processing and return */
@@ -776,7 +834,7 @@ skipp_real:
 		}
 	}
 
-	if (hw_polled)
+	if (likely(hw_polled))
 		bnxt_re_ring_cq_db(cq);
 
 	return dqed;
@@ -900,7 +958,6 @@ static int bnxt_re_poll_resize_cq_list(struct bnxt_re_cq *cq, uint32_t nwc,
 int bnxt_re_poll_cq(struct ibv_cq *ibvcq, int nwc, struct ibv_wc *wc)
 {
 	struct bnxt_re_cq *cq = to_bnxt_re_cq(ibvcq);
-	struct bnxt_re_context *cntx = to_bnxt_re_context(ibvcq->context);
 	int dqed = 0, left = 0;
 	uint32_t resize = 0;
 
@@ -909,7 +966,7 @@ int bnxt_re_poll_cq(struct ibv_cq *ibvcq, int nwc, struct ibv_wc *wc)
 	/* Check  whether we have anything to be completed
 	 * from prev cq context.
 	 */
-	if (!list_empty(&cq->prev_cq_head)) {
+	if (unlikely(!list_empty(&cq->prev_cq_head))) {
 		dqed = bnxt_re_poll_resize_cq_list(cq, nwc, wc);
 		left = nwc - dqed;
 		if (!left) {
@@ -918,21 +975,18 @@ int bnxt_re_poll_cq(struct ibv_cq *ibvcq, int nwc, struct ibv_wc *wc)
 		}
 	}
 	dqed += bnxt_re_poll_one(cq, left, wc + dqed, &resize);
-	pthread_spin_unlock(&cq->cqq.qlock);
 	left = nwc - dqed;
-	if (left) {
+	if (unlikely(left && (!list_empty(&cq->sfhead) ||
+			      !list_empty(&cq->rfhead))))
 		/* Check if anything is there to flush. */
-		pthread_spin_lock(&cntx->fqlock);
 		dqed += bnxt_re_poll_flush_lists(cq, left, (wc + dqed));
-		pthread_spin_unlock(&cntx->fqlock);
-	}
+	pthread_spin_unlock(&cq->cqq.qlock);
 
 	return dqed;
 }
 
 static void bnxt_re_cleanup_cq(struct bnxt_re_qp *qp, struct bnxt_re_cq *cq)
 {
-	struct bnxt_re_context *cntx;
 	struct bnxt_re_queue *que = &cq->cqq;
 	struct bnxt_re_bcqe *hdr;
 	struct bnxt_re_req_cqe *scqe;
@@ -940,8 +994,6 @@ static void bnxt_re_cleanup_cq(struct bnxt_re_qp *qp, struct bnxt_re_cq *cq)
 	void *cqe;
 	int indx, type;
 
-	cntx = to_bnxt_re_context(cq->ibvcq.context);
-
 	pthread_spin_lock(&que->qlock);
 	for (indx = 0; indx < que->depth; indx++) {
 		cqe = que->va + indx * bnxt_re_get_cqe_sz();
@@ -963,12 +1015,10 @@ static void bnxt_re_cleanup_cq(struct bnxt_re_qp *qp, struct bnxt_re_cq *cq)
 		}
 
 	}
-	pthread_spin_unlock(&que->qlock);
 
-	pthread_spin_lock(&cntx->fqlock);
 	bnxt_re_fque_del_node(&qp->snode);
 	bnxt_re_fque_del_node(&qp->rnode);
-	pthread_spin_unlock(&cntx->fqlock);
+	pthread_spin_unlock(&que->qlock);
 }
 
 int bnxt_re_arm_cq(struct ibv_cq *ibvcq, int flags)
@@ -1000,9 +1050,9 @@ static int bnxt_re_check_qp_limits(struct bnxt_re_context *cntx,
 	if (attr->cap.max_inline_data > BNXT_RE_MAX_INLINE_SIZE)
 		return EINVAL;
 	if (attr->cap.max_send_wr > devattr->max_qp_wr)
-		attr->cap.max_send_wr = devattr->max_qp_wr;
+		return EINVAL;
 	if (attr->cap.max_recv_wr > devattr->max_qp_wr)
-		attr->cap.max_recv_wr = devattr->max_qp_wr;
+		return EINVAL;
 
 	return 0;
 }
@@ -1033,15 +1083,13 @@ static int bnxt_re_alloc_queue_ptr(struct bnxt_re_qp *qp,
 
 	if (!attr->srq) {
 		qp->jrqq = calloc(1, sizeof(struct bnxt_re_joint_queue));
-		if (!qp->jrqq) {
-			free(qp->jsqq);
+		if (!qp->jrqq)
 			goto fail;
-		}
+
 		qp->jrqq->hwque = calloc(1, sizeof(struct bnxt_re_queue));
 		if (!qp->jrqq->hwque)
 			goto fail;
 	}
-
 	return 0;
 fail:
 	bnxt_re_free_queue_ptr(qp);
@@ -1127,13 +1175,13 @@ static int bnxt_re_get_sq_slots(struct bnxt_re_dev *rdev,
 	hdr_sz = bnxt_re_get_sqe_hdr_sz();
 	stride = sizeof(struct bnxt_re_sge);
 	max_wqesz = bnxt_re_calc_wqe_sz(rdev->devattr.max_sge);
-	ilsize = get_aligned(*ils, hdr_sz);
+	ilsize = align(*ils, hdr_sz);
 
 	wqe_size = bnxt_re_calc_wqe_sz(nsge);
 	if (ilsize) {
 		cal_ils = hdr_sz + ilsize;
 		wqe_size = MAX(cal_ils, wqe_size);
-		wqe_size = get_aligned(wqe_size, hdr_sz);
+		wqe_size = align(wqe_size, hdr_sz);
 	}
 	if (wqe_size > max_wqesz)
 		return -EINVAL;
@@ -1150,10 +1198,11 @@ static int bnxt_re_get_sq_slots(struct bnxt_re_dev *rdev,
 	return slots;
 }
 
-static int bnxt_re_alloc_queues(struct bnxt_re_dev *dev,
+static int bnxt_re_alloc_queues(struct bnxt_re_context *cntx,
 				struct bnxt_re_qp *qp,
 				struct ibv_qp_init_attr *attr,
-				uint32_t pg_size) {
+				uint32_t pg_size)
+{
 	struct bnxt_re_psns_ext *psns_ext;
 	struct bnxt_re_wrid *swque;
 	struct bnxt_re_queue *que;
@@ -1168,11 +1217,11 @@ static int bnxt_re_alloc_queues(struct bnxt_re_dev *dev,
 	que = qp->jsqq->hwque;
 	diff = (qp->qpmode == BNXT_RE_WQE_MODE_VARIABLE) ?
 		0 : BNXT_RE_FULL_FLAG_DELTA;
-	nswr = roundup_pow_of_two(attr->cap.max_send_wr + 1 + diff);
+	nswr = bnxt_re_init_depth(attr->cap.max_send_wr + 1 + diff, cntx->comp_mask);
 	nsge = attr->cap.max_send_sge;
 	if (nsge % 2)
 		nsge++;
-	nslots = bnxt_re_get_sq_slots(dev, qp, nswr, nsge,
+	nslots = bnxt_re_get_sq_slots(cntx->rdev, qp, nswr, nsge,
 				      &attr->cap.max_inline_data);
 	if (nslots < 0)
 		 return nslots;
@@ -1180,10 +1229,11 @@ static int bnxt_re_alloc_queues(struct bnxt_re_dev *dev,
 	que->depth = nslots;
 	que->diff = (diff * que->esize) / que->stride;
 
+	que->pad = (que->va + que->depth * que->stride);
 	/* psn_depth extra entries of size que->stride */
-	psn_size = qp->cctx->gen_p5 ? sizeof(struct bnxt_re_psns_ext) :
-				      sizeof(struct bnxt_re_psns);
+	psn_size = bnxt_re_get_psne_size(qp->cntx);
 	psn_depth = (nswr * psn_size) / que->stride;
+	que->pad_stride_log2 = (uint32_t)ilog32(psn_size);
 	if ((nswr * psn_size) % que->stride)
 		psn_depth++;
 	que->depth += psn_depth;
@@ -1210,12 +1260,23 @@ static int bnxt_re_alloc_queues(struct bnxt_re_dev *dev,
 	swque = qp->jsqq->swque;
 	for (indx = 0 ; indx < nswr; indx++, psns++)
 		swque[indx].psns = psns;
-	if (qp->cctx->gen_p5) {
+	if (qp->cctx->gen_p5_p7) {
 		for (indx = 0 ; indx < nswr; indx++, psns_ext++) {
 			swque[indx].psns_ext = psns_ext;
 			swque[indx].psns = (struct bnxt_re_psns *)psns_ext;
 		}
 	}
+	/* Init and adjust MSN table size according to qp mode */
+	if (!BNXT_RE_HW_RETX(qp->cntx))
+		goto skip_msn;
+	que->msn = 0;
+	que->msn_tbl_sz = 0;
+	if (qp->qpmode & BNXT_RE_WQE_MODE_VARIABLE)
+		que->msn_tbl_sz = roundup_pow_of_two(nslots) / 2;
+	else
+		que->msn_tbl_sz = roundup_pow_of_two(nswr);
+skip_msn:
+
 	qp->cap.max_swr = nswr;
 	pthread_spin_init(&que->qlock, PTHREAD_PROCESS_PRIVATE);
 
@@ -1224,11 +1285,11 @@ static int bnxt_re_alloc_queues(struct bnxt_re_dev *dev,
 
 	if (qp->jrqq) {
 		que = qp->jrqq->hwque;
-		nswr = roundup_pow_of_two(attr->cap.max_recv_wr + 1);
+		nswr = bnxt_re_init_depth(attr->cap.max_recv_wr + 1, cntx->comp_mask);
 		nsge = attr->cap.max_recv_sge;
 		if (nsge % 2)
 			nsge++;
-		nslots = bnxt_re_get_rq_slots(dev, qp, nswr, nsge);
+		nslots = bnxt_re_get_rq_slots(cntx->rdev, qp, nswr, nsge);
 		if (nslots < 0) {
 			ret = nslots;
 			goto fail;
@@ -1255,6 +1316,37 @@ fail:
 	return ret;
 }
 
+void bnxt_re_async_event(struct ibv_context *context,
+		      struct ibv_async_event *event)
+{
+	struct ibv_qp *ibvqp;
+	struct bnxt_re_qp *qp;
+
+	switch (event->event_type) {
+	case IBV_EVENT_CQ_ERR:
+		break;
+	case IBV_EVENT_SRQ_ERR:
+	case IBV_EVENT_QP_FATAL:
+	case IBV_EVENT_QP_REQ_ERR:
+	case IBV_EVENT_QP_ACCESS_ERR:
+	case IBV_EVENT_PATH_MIG_ERR: {
+		ibvqp = event->element.qp;
+		qp = to_bnxt_re_qp(ibvqp);
+		bnxt_re_qp_move_flush_err(qp);
+		break;
+	}
+	case IBV_EVENT_SQ_DRAINED:
+	case IBV_EVENT_PATH_MIG:
+	case IBV_EVENT_COMM_EST:
+	case IBV_EVENT_QP_LAST_WQE_REACHED:
+	case IBV_EVENT_SRQ_LIMIT_REACHED:
+	case IBV_EVENT_PORT_ACTIVE:
+	case IBV_EVENT_PORT_ERR:
+	default:
+		break;
+	}
+}
+
 struct ibv_qp *bnxt_re_create_qp(struct ibv_pd *ibvpd,
 				 struct ibv_qp_init_attr *attr)
 {
@@ -1279,7 +1371,7 @@ struct ibv_qp *bnxt_re_create_qp(struct ibv_pd *ibvpd,
 	qp->cctx = &cntx->cctx;
 	qp->qpmode = cntx->wqe_mode & BNXT_RE_WQE_MODE_VARIABLE;
 	qp->cntx = cntx;
-	if (bnxt_re_alloc_queues(dev, qp, attr, dev->pg_size))
+	if (bnxt_re_alloc_queues(cntx, qp, attr, dev->pg_size))
 		goto failq;
 	/* Fill ibv_cmd */
 	cap = &qp->cap;
@@ -1300,15 +1392,17 @@ struct ibv_qp *bnxt_re_create_qp(struct ibv_pd *ibvpd,
 	if (attr->srq)
 		qp->srq = to_bnxt_re_srq(attr->srq);
 	qp->udpi = &cntx->udpi;
+	qp->rand.seed = qp->qpid;
 	/* Save/return the altered Caps. */
 	cap->max_ssge = attr->cap.max_send_sge;
 	cap->max_rsge = attr->cap.max_recv_sge;
 	cap->max_inline = attr->cap.max_inline_data;
 	cap->sqsig = attr->sq_sig_all;
+	cap->is_atomic_cap = dev->devattr.atomic_cap;
 	fque_init_node(&qp->snode);
 	fque_init_node(&qp->rnode);
 
-	if (qp->cctx->gen_p5 && cntx->udpi.wcdpi) {
+	if (qp->cctx->gen_p5_p7 && cntx->udpi.wcdpi) {
 		qp->push_st_en = 1;
 		qp->max_push_sz = BNXT_RE_MAX_INLINE_SIZE;
 	}
@@ -1395,8 +1489,22 @@ int bnxt_re_destroy_qp(struct ibv_qp *ibvqp)
 	return 0;
 }
 
-static int bnxt_re_put_sge(struct bnxt_re_queue *que, uint32_t *idx,
-			   struct ibv_sge *sgl, int nsg)
+static void bnxt_re_put_rx_sge(struct bnxt_re_queue *que, uint32_t *idx,
+			       struct ibv_sge *sgl, int nsg)
+{
+	struct bnxt_re_sge *sge;
+	int indx;
+
+	for (indx = 0; indx < nsg; indx++) {
+		sge = bnxt_re_get_hwqe(que, (*idx)++);
+		sge->pa = htole64(sgl[indx].addr);
+		sge->lkey = htole32(sgl[indx].lkey);
+		sge->length = htole32(sgl[indx].length);
+	}
+}
+
+static int bnxt_re_put_tx_sge(struct bnxt_re_queue *que, uint32_t *idx,
+			      struct ibv_sge *sgl, int nsg)
 {
 	struct bnxt_re_sge *sge;
 	int indx;
@@ -1420,7 +1528,7 @@ static inline int bnxt_re_calc_inline_len(struct ibv_send_wr *swr)
 	illen = 0;
 	for (indx = 0; indx < swr->num_sge; indx++)
 		illen += swr->sg_list[indx].length;
-	return get_aligned(illen, sizeof(struct bnxt_re_sge));
+	return align(illen, sizeof(struct bnxt_re_sge));
 }
 
 static int bnxt_re_put_inline(struct bnxt_re_queue *que, uint32_t *idx,
@@ -1475,7 +1583,7 @@ bad:
 }
 
 static int bnxt_re_required_slots(struct bnxt_re_qp *qp, struct ibv_send_wr *wr,
-				  uint32_t *wqe_sz, bool *push)
+				  uint32_t *wqe_sz, void **pbuf)
 {
 	uint32_t wqe_byte;
 	int ilsize;
@@ -1485,7 +1593,7 @@ static int bnxt_re_required_slots(struct bnxt_re_qp *qp, struct ibv_send_wr *wr,
 		if (ilsize > qp->cap.max_inline)
 			return -EINVAL;
 		if (qp->push_st_en && ilsize <= qp->max_push_sz)
-			*push = true;
+			*pbuf = bnxt_re_get_pbuf(&qp->push_st_en, qp->cntx);
 		wqe_byte = (ilsize + bnxt_re_get_sqe_hdr_sz());
 	} else {
 		wqe_byte = bnxt_re_calc_wqe_sz(wr->num_sge);
@@ -1528,23 +1636,62 @@ static inline void bnxt_re_set_hdr_flags(struct bnxt_re_bsqe *hdr,
 	hdr->rsv_ws_fl_wt = htole32(hdrval);
 }
 
-static int bnxt_re_build_sge(struct bnxt_re_queue *que, uint32_t *idx,
-			     struct bnxt_re_push_buffer *pbuf,
-			     struct ibv_send_wr *wr,
-			     uint16_t max_il)
+static int bnxt_re_build_tx_sge(struct bnxt_re_queue *que, uint32_t *idx,
+				struct bnxt_re_push_buffer *pbuf,
+				struct ibv_send_wr *wr,
+				uint16_t max_il)
 {
 	if (wr->send_flags & IBV_SEND_INLINE)
 		return bnxt_re_put_inline(que, idx, pbuf, wr->sg_list, wr->num_sge, max_il);
 
-	return bnxt_re_put_sge(que, idx, wr->sg_list, wr->num_sge);
+	return bnxt_re_put_tx_sge(que, idx, wr->sg_list, wr->num_sge);
+}
+
+static void *bnxt_re_pull_psn_buff(struct bnxt_re_queue *que, bool hw_retx)
+{
+	if (hw_retx)
+		return (void *)(que->pad + ((que->msn) << que->pad_stride_log2));
+	return (void *)(que->pad + ((*que->dbtail) << que->pad_stride_log2));
+}
+
+static void bnxt_re_fill_psns_for_msntbl(struct bnxt_re_qp *qp, uint32_t len, uint32_t st_idx)
+{
+	uint32_t npsn = 0, start_psn = 0, next_psn = 0;
+	struct bnxt_re_msns *msns;
+	uint32_t pkt_cnt = 0;
+
+	msns = bnxt_re_pull_psn_buff(qp->jsqq->hwque, true);
+	msns->start_idx_next_psn_start_psn = 0;
+
+	if (qp->qptyp == IBV_QPT_RC) {
+		start_psn = qp->sq_psn;
+		pkt_cnt = (len / qp->mtu);
+		if (len % qp->mtu)
+			pkt_cnt++;
+		/* Increment the psn even for 0 len packets
+		 * e.g. for opcode rdma-write-with-imm-data
+		 * with length field = 0
+		 */
+		if (len == 0)
+			pkt_cnt = 1;
+		/* make it 24 bit */
+		next_psn = qp->sq_psn + pkt_cnt;
+		npsn = next_psn;
+		qp->sq_psn = next_psn;
+		msns->start_idx_next_psn_start_psn |=
+			bnxt_re_update_msn_tbl(st_idx, npsn, start_psn);
+		qp->jsqq->hwque->msn++;
+		qp->jsqq->hwque->msn %= qp->jsqq->hwque->msn_tbl_sz;
+	}
 }
 
 static void bnxt_re_fill_psns(struct bnxt_re_qp *qp, struct bnxt_re_wrid *wrid,
-			      uint8_t opcode, uint32_t len)
+			      uint32_t len)
 {
 	uint32_t opc_spsn = 0, flg_npsn = 0;
 	struct bnxt_re_psns_ext *psns_ext;
 	uint32_t pkt_cnt = 0, nxt_psn = 0;
+	uint8_t opcode = wrid->wc_opcd;
 	struct bnxt_re_psns *psns;
 
 	psns = wrid->psns;
@@ -1562,13 +1709,12 @@ static void bnxt_re_fill_psns(struct bnxt_re_qp *qp, struct bnxt_re_wrid *wrid,
 		flg_npsn = nxt_psn;
 		qp->sq_psn = nxt_psn;
 	}
-	opcode = bnxt_re_ibv_wr_to_wc_opcd(opcode);
 	opc_spsn |= (((uint32_t)opcode & BNXT_RE_PSNS_OPCD_MASK) <<
 		      BNXT_RE_PSNS_OPCD_SHIFT);
 	memset(psns, 0, sizeof(*psns));
 	psns->opc_spsn = htole32(opc_spsn);
 	psns->flg_npsn = htole32(flg_npsn);
-	if (qp->cctx->gen_p5)
+	if (qp->cctx->gen_p5_p7)
 		psns_ext->st_slot_idx = wrid->st_slot_idx;
 }
 
@@ -1603,6 +1749,12 @@ static int bnxt_re_build_ud_sqe(struct ibv_send_wr *wr,
 	return 0;
 }
 
+static bool __atomic_not_supported(struct bnxt_re_qp *qp, struct ibv_send_wr *wr)
+{
+	/* Atomic capability disabled or the request has more than 1 SGE */
+	return (!qp->cap.is_atomic_cap || wr->num_sge > 1);
+}
+
 static void bnxt_re_build_cns_sqe(struct ibv_send_wr *wr,
 				  struct bnxt_re_bsqe *hdr,
 				  void *hdr2)
@@ -1626,6 +1778,36 @@ static void bnxt_re_build_fna_sqe(struct ibv_send_wr *wr,
 	sqe->swp_dt = htole64(wr->wr.atomic.compare_add);
 }
 
+static int bnxt_re_build_atomic_sqe(struct bnxt_re_qp *qp,
+				    struct ibv_send_wr *wr,
+				    struct bnxt_re_bsqe *hdr,
+				    void *hdr2)
+{
+	if (__atomic_not_supported(qp, wr))
+		return -EINVAL;
+	switch (wr->opcode) {
+	case IBV_WR_ATOMIC_CMP_AND_SWP:
+		bnxt_re_build_cns_sqe(wr, hdr, hdr2);
+		return 0;
+	case IBV_WR_ATOMIC_FETCH_AND_ADD:
+		bnxt_re_build_fna_sqe(wr, hdr, hdr2);
+		return 0;
+	default:
+		return -EINVAL;
+	}
+}
+
+static void bnxt_re_force_rts2rts(struct bnxt_re_qp *qp)
+{
+	struct ibv_qp_attr attr;
+	int attr_mask;
+
+	attr_mask = IBV_QP_STATE;
+	attr.qp_state = IBV_QPS_RTS;
+	bnxt_re_modify_qp(&qp->ibvqp, &attr, attr_mask);
+	qp->wqe_cnt = 0;
+}
+
 int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 		      struct ibv_send_wr **bad)
 {
@@ -1641,15 +1823,13 @@ int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 	uint32_t wqe_size = 0;
 	bool ring_db = false;
 	uint8_t sig = 0;
-	bool try_push;
 	uint32_t idx;
 
 	pthread_spin_lock(&sq->qlock);
 	while (wr) {
 
-		try_push = false;
 		pbuf = NULL;
-		slots = bnxt_re_required_slots(qp, wr, &wqe_size, &try_push);
+		slots = bnxt_re_required_slots(qp, wr, &wqe_size, (void **)&pbuf);
 		if (bnxt_re_is_que_full(sq, slots) ||
 		    wr->num_sge > qp->cap.max_ssge) {
 			*bad = wr;
@@ -1657,20 +1837,21 @@ int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 			goto bad_wr;
 		}
 
-		idx = 0;
-		hdr = bnxt_re_get_hwqe(sq, idx++);
-		sqe = bnxt_re_get_hwqe(sq, idx++);
-		if (try_push) {
-			pbuf = bnxt_re_get_pbuf(&qp->push_st_en, qp->cntx);
-			if (pbuf) {
-				pbuf->qpid = qp->qpid;
-				pbuf->wqe[0] = (uintptr_t)hdr;
-				pbuf->wqe[1] = (uintptr_t)sqe;
-				pbuf->st_idx = *sq->dbtail;
-			}
+		idx = 2;
+		bytes = 0;
+		hdr = bnxt_re_get_hwqe(sq, 0);
+		sqe = bnxt_re_get_hwqe(sq, 1);
+
+		/* populate push buffer */
+		if (pbuf) {
+			pbuf->qpid = qp->qpid;
+			pbuf->wqe[0] = (uintptr_t)hdr;
+			pbuf->wqe[1] = (uintptr_t)sqe;
+			pbuf->st_idx = *sq->dbtail;
 		}
+
 		if (wr->num_sge) {
-			bytes = bnxt_re_build_sge(sq, &idx, pbuf, wr, qp->cap.max_inline);
+			bytes = bnxt_re_build_tx_sge(sq, &idx, pbuf, wr, qp->cap.max_inline);
 			if (unlikely(bytes < 0)) {
 				ret = ENOMEM;
 				*bad = wr;
@@ -1708,10 +1889,12 @@ int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 			rsqe->rkey = htole32(wr->wr.rdma.rkey);
 			break;
 		case IBV_WR_ATOMIC_CMP_AND_SWP:
-			bnxt_re_build_cns_sqe(wr, hdr, sqe);
-			break;
 		case IBV_WR_ATOMIC_FETCH_AND_ADD:
-			bnxt_re_build_fna_sqe(wr, hdr, sqe);
+			if (bnxt_re_build_atomic_sqe(qp, wr, hdr, sqe)) {
+				ret = EINVAL;
+				*bad = wr;
+				goto bad_wr;
+			}
 			break;
 		default:
 			ret = -EINVAL;
@@ -1723,7 +1906,11 @@ int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 		sig = ((wr->send_flags & IBV_SEND_SIGNALED) || qp->cap.sqsig);
 		bnxt_re_fill_wrid(wrid, wr->wr_id, bytes,
 				  sig, sq->tail, slots);
-		bnxt_re_fill_psns(qp, wrid, wr->opcode, bytes);
+		wrid->wc_opcd = bnxt_re_ibv_wr_to_wc_opcd(wr->opcode);
+		if (BNXT_RE_HW_RETX(qp->cntx))
+			bnxt_re_fill_psns_for_msntbl(qp, bytes, *sq->dbtail);
+		else
+			bnxt_re_fill_psns(qp, wrid, bytes);
 		bnxt_re_jqq_mod_start(qp->jsqq, swq_idx);
 		bnxt_re_incr_tail(sq, slots);
 		ring_db = true;
@@ -1733,28 +1920,24 @@ int bnxt_re_post_send(struct ibv_qp *ibvqp, struct ibv_send_wr *wr,
 			pbuf->tail = *sq->dbtail;
 			bnxt_re_fill_push_wcb(qp, pbuf, idx);
 			bnxt_re_put_pbuf(qp->cntx, pbuf);
+			pbuf = NULL;
 		}
 
 		qp->wqe_cnt++;
 		wr = wr->next;
 
-		if (qp->wqe_cnt == BNXT_RE_UD_QP_HW_STALL &&
-		    qp->qptyp == IBV_QPT_UD) {
-			/* Move RTS to RTS since it is time. */
-			struct ibv_qp_attr attr;
-			int attr_mask;
-
-			attr_mask = IBV_QP_STATE;
-			attr.qp_state = IBV_QPS_RTS;
-			bnxt_re_modify_qp(&qp->ibvqp, &attr, attr_mask);
-			qp->wqe_cnt = 0;
-		}
+		if (unlikely(!qp->cntx->cctx.gen_p5_p7 && qp->wqe_cnt == BNXT_RE_UD_QP_HW_STALL &&
+			     qp->qptyp == IBV_QPT_UD))
+			bnxt_re_force_rts2rts(qp);
 	}
 
 bad_wr:
 	if (ring_db)
 		bnxt_re_ring_sq_db(qp);
 
+	if (pbuf)
+		bnxt_re_put_pbuf(qp->cntx, pbuf);
+
 	pthread_spin_unlock(&sq->qlock);
 	return ret;
 }
@@ -1768,10 +1951,10 @@ int bnxt_re_post_recv(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
 	struct bnxt_re_brqe *hdr;
 	struct bnxt_re_sge *sge;
 	bool ring_db = false;
-	int ret = 0, rc = 0;
 	uint32_t hdrval = 0;
 	uint32_t idx = 0;
 	uint32_t swq_idx;
+	int rc = 0;
 
 	pthread_spin_lock(&rq->qlock);
 	while (wr) {
@@ -1790,7 +1973,7 @@ int bnxt_re_post_recv(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
 		idx = 2;
 		hdr = bnxt_re_get_hwqe_hdr(rq);
 
-		if (!wr->num_sge) {
+		if (unlikely(!wr->num_sge)) {
 			/*
 			 * HW needs at least one SGE for RQ Entries.
 			 * Create an entry if num_sge = 0,
@@ -1800,12 +1983,7 @@ int bnxt_re_post_recv(struct ibv_qp *ibvqp, struct ibv_recv_wr *wr,
 			sge->length = 0;
 		} else {
 			/* Fill SGEs */
-			ret = bnxt_re_put_sge(rq, &idx, wr->sg_list, wr->num_sge);
-		}
-		if (unlikely(ret < 0)) {
-			*bad = wr;
-			rc = ENOMEM;
-			break;
+			bnxt_re_put_rx_sge(rq, &idx, wr->sg_list, wr->num_sge);
 		}
 		hdrval = BNXT_RE_WR_OPCD_RECV;
 		hdrval |= ((idx & BNXT_RE_HDR_WS_MASK) << BNXT_RE_HDR_WS_SHIFT);
@@ -1858,7 +2036,8 @@ static void bnxt_re_srq_free_queue(struct bnxt_re_srq *srq)
 	bnxt_re_free_aligned(srq->srqq);
 }
 
-static int bnxt_re_srq_alloc_queue(struct bnxt_re_srq *srq,
+static int bnxt_re_srq_alloc_queue(struct bnxt_re_context *cntx,
+				   struct bnxt_re_srq *srq,
 				   struct ibv_srq_init_attr *attr,
 				   uint32_t pg_size)
 {
@@ -1866,7 +2045,7 @@ static int bnxt_re_srq_alloc_queue(struct bnxt_re_srq *srq,
 	int ret, idx;
 
 	que = srq->srqq;
-	que->depth = roundup_pow_of_two(attr->attr.max_wr + 1);
+	que->depth = bnxt_re_init_depth(attr->attr.max_wr + 1, cntx->comp_mask);
 	que->diff = que->depth - attr->attr.max_wr;
 	que->stride = bnxt_re_get_srqe_sz();
 	ret = bnxt_re_alloc_aligned(que, pg_size);
@@ -1908,7 +2087,7 @@ struct ibv_srq *bnxt_re_create_srq(struct ibv_pd *ibvpd,
 	if (!srq)
 		goto fail;
 
-	if (bnxt_re_srq_alloc_queue(srq, attr, dev->pg_size))
+	if (bnxt_re_srq_alloc_queue(cntx, srq, attr, dev->pg_size))
 		goto fail;
 
 	req.srqva = (uintptr_t)srq->srqq->va;
@@ -1920,7 +2099,10 @@ struct ibv_srq *bnxt_re_create_srq(struct ibv_pd *ibvpd,
 		goto fail;
 
 	srq->srqid = resp.srqid;
+	srq->cntx = cntx;
 	srq->udpi = &cntx->udpi;
+	srq->rand.seed = srq->srqid;
+
 	srq->cap.max_wr = srq->srqq->depth;
 	srq->cap.max_sge = attr->attr.max_sge;
 	srq->cap.srq_limit = attr->attr.srq_limit;
