@@ -39,6 +39,7 @@
 #include <linux/types.h>
 #include <util/compiler.h>
 #include <util/udma_barrier.h>
+#include <util/util.h>
 #include <endian.h>
 
 /*
@@ -53,7 +54,6 @@
 #define __iomem
 #define BUG_ON(c) assert(!(c))
 #define ROUND_UP(x, n) (((x) + (n) - 1u) & ~((n) - 1u))
-#define DIV_ROUND_UP(n,d) (((n) + (d) - 1) / (d))
 
 /* FIXME: Move me to a generic PCI mmio accessor */
 #define cpu_to_pci32(val) htole32(val)
