@@ -60,7 +60,6 @@ enum {
 	HNS_ROCE_WQE_OP_ATOMIC_MASK_COMP_AND_SWAP = 0x8,
 	HNS_ROCE_WQE_OP_ATOMIC_MASK_FETCH_AND_ADD = 0x9,
 	HNS_ROCE_WQE_OP_FAST_REG_PMR = 0xa,
-	HNS_ROCE_WQE_OP_LOCAL_INV = 0xb,
 	HNS_ROCE_WQE_OP_BIND_MW_TYPE = 0xc,
 	HNS_ROCE_WQE_OP_MASK = 0x1f
 };
@@ -85,7 +84,6 @@ enum {
 	HNS_ROCE_SQ_OP_ATOMIC_MASK_COMP_AND_SWAP = 0x8,
 	HNS_ROCE_SQ_OP_ATOMIC_MASK_FETCH_AND_ADD = 0x9,
 	HNS_ROCE_SQ_OP_FAST_REG_PMR = 0xa,
-	HNS_ROCE_SQ_OP_LOCAL_INV = 0xb,
 	HNS_ROCE_SQ_OP_BIND_MW = 0xc,
 };
 
@@ -110,6 +108,7 @@ enum {
 	HNS_ROCE_V2_CQE_TRANSPORT_RETRY_EXC_ERR		= 0x15,
 	HNS_ROCE_V2_CQE_RNR_RETRY_EXC_ERR		= 0x16,
 	HNS_ROCE_V2_CQE_REMOTE_ABORTED_ERR		= 0x22,
+	HNS_ROCE_V2_CQE_GENERAL_ERR			= 0x23,
 	HNS_ROCE_V2_CQE_XRC_VIOLATION_ERR		= 0x24,
 };
 
@@ -121,6 +120,11 @@ enum {
 	HNS_ROCE_V2_CQ_DB_NTR,
 };
 
+enum hns_roce_wr_buf_type {
+	WR_BUF_TYPE_POST_SEND,
+	WR_BUF_TYPE_SEND_WR_OPS,
+};
+
 struct hns_roce_db {
 	__le32	byte_4;
 	__le32	parameter;
@@ -151,7 +155,7 @@ struct hns_roce_v2_cqe {
 	__le32	smac;
 	__le32	byte_28;
 	__le32	byte_32;
-	__le32	rsv[8];
+	__le32	payload[8];
 };
 
 #define CQE_FIELD_LOC(h, l) FIELD_LOC(struct hns_roce_v2_cqe, h, l)
@@ -164,7 +168,7 @@ struct hns_roce_v2_cqe {
 #define CQE_WQE_IDX CQE_FIELD_LOC(31, 16)
 #define CQE_RKEY_IMMTDATA CQE_FIELD_LOC(63, 32)
 #define CQE_XRC_SRQN CQE_FIELD_LOC(87, 64)
-#define CQE_RSV0 CQE_FIELD_LOC(95, 88)
+#define CQE_CQE_INLINE CQE_FIELD_LOC(89, 88)
 #define CQE_LCL_QPN CQE_FIELD_LOC(119, 96)
 #define CQE_SUB_STATUS CQE_FIELD_LOC(127, 120)
 #define CQE_BYTE_CNT CQE_FIELD_LOC(159, 128)
@@ -337,5 +341,8 @@ struct hns_roce_ud_sq_wqe {
 #define MAX_SERVICE_LEVEL 0x7
 
 void hns_roce_v2_clear_qp(struct hns_roce_context *ctx, struct hns_roce_qp *qp);
+void hns_roce_attach_cq_ex_ops(struct ibv_cq_ex *cq_ex, uint64_t wc_flags);
+int hns_roce_attach_qp_ex_ops(struct ibv_qp_init_attr_ex *attr,
+			      struct hns_roce_qp *qp);
 
 #endif /* _HNS_ROCE_U_HW_V2_H */
