@@ -422,6 +422,7 @@ struct mlx5_context {
 	pthread_mutex_t			crypto_login_mutex;
 	uint64_t			max_dc_rd_atom;
 	uint64_t			max_dc_init_rd_atom;
+	struct mlx5dv_reg		reg_c0;
 };
 
 struct mlx5_hugetlb_mem {
@@ -560,6 +561,11 @@ struct mlx5_srq {
 	int				op_tail;
 	int				unexp_in;
 	int				unexp_out;
+	/* Bit is set if WQE is in SW ownership and not part of the SRQ queues
+	 * (main/wait)
+	 */
+	unsigned long *free_wqe_bitmap;
+	uint32_t nwqes;
 };
 
 
