@@ -35,7 +35,7 @@
 
 #include <ccan/list.h>
 #include <ccan/minmax.h>
-#include <ccan/bitmap.h>
+#include <util/bitmap.h>
 #include <stdatomic.h>
 #include "mlx5dv.h"
 #include "mlx5_ifc.h"
@@ -57,6 +57,7 @@
 #define DR_STE_MAX_FLEX_0_ID	3
 #define DR_STE_MAX_FLEX_1_ID	7
 #define DR_VPORTS_BUCKETS	256
+#define ACTION_CACHE_LINE_SIZE	64
 
 #define dr_dbg(dmn, arg...) dr_dbg_ctx((dmn)->ctx, ##arg)
 
@@ -87,12 +88,18 @@ enum dr_icm_chunk_size {
 	DR_CHUNK_SIZE_512K,
 	DR_CHUNK_SIZE_1024K,
 	DR_CHUNK_SIZE_2048K,
+	DR_CHUNK_SIZE_4096K,
+	DR_CHUNK_SIZE_8192K,
+	DR_CHUNK_SIZE_16384K,
 	DR_CHUNK_SIZE_MAX,
 };
 
 enum dr_icm_type {
 	DR_ICM_TYPE_STE,
 	DR_ICM_TYPE_MODIFY_ACTION,
+	DR_ICM_TYPE_MODIFY_HDR_PTRN,
+	DR_ICM_TYPE_ENCAP,
+	DR_ICM_TYPE_MAX,
 };
 
 static inline enum dr_icm_chunk_size
@@ -124,6 +131,8 @@ enum dr_ste_ctx_action_cap {
 	DR_STE_CTX_ACTION_CAP_TX_POP	= 1 << 0,
 	DR_STE_CTX_ACTION_CAP_RX_PUSH	= 1 << 1,
 	DR_STE_CTX_ACTION_CAP_RX_ENCAP	= 1 << 3,
+	DR_STE_CTX_ACTION_CAP_POP_MDFY	= 1 << 4,
+	DR_STE_CTX_ACTION_CAP_MODIFY_HDR_INLINE = 1 << 5,
 };
 
 enum {
@@ -131,6 +140,11 @@ enum {
 	DR_MODIFY_ACTION_LOG_SIZE	= 3,
 };
 
+enum {
+	DR_SW_ENCAP_ENTRY_SIZE		= 64,
+	DR_SW_ENCAP_ENTRY_LOG_SIZE	= 6,
+};
+
 enum dr_matcher_criteria {
 	DR_MATCHER_CRITERIA_EMPTY	= 0,
 	DR_MATCHER_CRITERIA_OUTER	= 1 << 0,
@@ -177,6 +191,7 @@ enum dr_action_type {
 	DR_ACTION_TYP_ASO_FIRST_HIT,
 	DR_ACTION_TYP_ASO_FLOW_METER,
 	DR_ACTION_TYP_ASO_CT,
+	DR_ACTION_TYP_ROOT_FT,
 	DR_ACTION_TYP_MAX,
 };
 
@@ -189,6 +204,10 @@ struct dr_devx_caps;
 struct dr_rule_rx_tx;
 struct dr_matcher_rx_tx;
 struct dr_ste_ctx;
+struct dr_ptrn_mngr;
+struct dr_ptrn_obj;
+struct dr_arg_mngr;
+struct dr_arg_obj;
 
 struct dr_data_seg {
 	uint64_t	addr;
@@ -197,7 +216,13 @@ struct dr_data_seg {
 	unsigned int	send_flags;
 };
 
+enum send_info_type {
+	WRITE_ICM = 0,
+	GTA_ARG   = 1,
+};
+
 struct postsend_info {
+	enum send_info_type	type;
 	struct dr_data_seg	write;
 	struct dr_data_seg	read;
 	uint64_t		remote_addr;
@@ -363,10 +388,15 @@ struct dr_action_aso {
 	};
 };
 
+#define DR_INVALID_PATTERN_INDEX 0xffffffff
+
 struct dr_ste_actions_attr {
 	uint32_t	modify_index;
+	uint32_t	modify_pat_idx;
 	uint16_t	modify_actions;
+	uint8_t		*single_modify_action;
 	uint32_t	decap_index;
+	uint32_t	decap_pat_idx;
 	uint16_t	decap_actions;
 	bool		decap_with_vlan;
 	uint64_t	final_icm_addr;
@@ -378,11 +408,13 @@ struct dr_ste_actions_attr {
 	uint32_t	reformat_size;
 	bool		prio_tag_required;
 	struct {
-		int		count;
+		int		count_pop;
+		int		count_push;
 		uint32_t	headers[MAX_VLANS];
 	} vlans;
 	struct dr_action_aso *aso;
 	uint32_t aso_ste_loc;
+	struct mlx5dv_dr_domain *dmn;
 };
 
 struct cross_dmn_params {
@@ -549,6 +581,11 @@ void dr_ste_build_tnl_geneve_tlv_opt(struct dr_ste_ctx *ste_ctx,
 				     struct dr_match_param *mask,
 				     struct dr_devx_caps *caps,
 				     bool inner, bool rx);
+void dr_ste_build_tnl_geneve_tlv_opt_exist(struct dr_ste_ctx *ste_ctx,
+					   struct dr_ste_build *sb,
+					   struct dr_match_param *mask,
+					   struct dr_devx_caps *caps,
+					   bool inner, bool rx);
 void dr_ste_build_tnl_gtpu(struct dr_ste_ctx *ste_ctx,
 			   struct dr_ste_build *sb,
 			   struct dr_match_param *mask,
@@ -588,10 +625,15 @@ void dr_ste_build_flex_parser_1(struct dr_ste_ctx *ste_ctx,
 				struct dr_ste_build *sb,
 				struct dr_match_param *mask,
 				bool inner, bool rx);
-void dr_ste_build_tunnel_header_0_1(struct dr_ste_ctx *ste_ctx,
-				    struct dr_ste_build *sb,
-				    struct dr_match_param *mask,
-				    bool inner, bool rx);
+void dr_ste_build_tunnel_header(struct dr_ste_ctx *ste_ctx,
+				struct dr_ste_build *sb,
+				struct dr_match_param *mask,
+				struct dr_devx_caps *caps,
+				bool inner, bool rx);
+void dr_ste_build_ib_l4(struct dr_ste_ctx *ste_ctx,
+			struct dr_ste_build *sb,
+			struct dr_match_param *mask,
+			bool inner, bool rx);
 int dr_ste_build_def0(struct dr_ste_ctx *ste_ctx,
 		      struct dr_ste_build *sb,
 		      struct dr_match_param *mask,
@@ -644,40 +686,44 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			     uint32_t num_actions,
 			     uint8_t *ste_arr,
 			     uint32_t *new_hw_ste_arr_sz,
-			     struct cross_dmn_params *cross_dmn_p);
+			     struct cross_dmn_params *cross_dmn_p,
+			     uint8_t send_ring_idx);
 int dr_actions_build_attr(struct mlx5dv_dr_matcher *matcher,
 			  struct mlx5dv_dr_action *actions[],
 			  size_t num_actions,
 			  struct mlx5dv_flow_action_attr *attr,
 			  struct mlx5_flow_action_attr_aux *attr_aux);
 
+uint32_t dr_actions_reformat_get_id(struct mlx5dv_dr_action *action);
+
 struct dr_match_spec {
 	uint32_t smac_47_16;	/* Source MAC address of incoming packet */
-	uint32_t ethertype:16;	/* Incoming packet Ethertype - this is the Ethertype following the last ;VLAN tag of the packet */
 	uint32_t smac_15_0:16;	/* Source MAC address of incoming packet */
+	uint32_t ethertype:16;	/* Incoming packet Ethertype - this is the Ethertype following the last ;VLAN tag of the packet */
 	uint32_t dmac_47_16;	/* Destination MAC address of incoming packet */
-	uint32_t first_vid:12;	/* VLAN ID of first VLAN tag in the incoming packet. Valid only ;when cvlan_tag==1 or svlan_tag==1 */
-	uint32_t first_cfi:1;	/* CFI bit of first VLAN tag in the incoming packet. Valid only when ;cvlan_tag==1 or svlan_tag==1 */
-	uint32_t first_prio:3;	/* Priority of first VLAN tag in the incoming packet. Valid only when ;cvlan_tag==1 or svlan_tag==1 */
 	uint32_t dmac_15_0:16;	/* Destination MAC address of incoming packet */
-	uint32_t tcp_flags:9;	/* TCP flags. ;Bit 0: FIN;Bit 1: SYN;Bit 2: RST;Bit 3: PSH;Bit 4: ACK;Bit 5: URG;Bit 6: ECE;Bit 7: CWR;Bit 8: NS */
-	uint32_t ip_version:4;	/* IP version */
-	uint32_t frag:1;	/* Packet is an IP fragment */
-	uint32_t svlan_tag:1;	/* The first vlan in the packet is s-vlan (0x8a88). cvlan_tag and ;svlan_tag cannot be set together */
-	uint32_t cvlan_tag:1;	/* The first vlan in the packet is c-vlan (0x8100). cvlan_tag and ;svlan_tag cannot be set together */
-	uint32_t ip_ecn:2;	/* Explicit Congestion Notification derived from Traffic Class/TOS ;field of IPv6/v4 */
-	uint32_t ip_dscp:6;	/* Differentiated Services Code Point derived from Traffic Class/;TOS field of IPv6/v4 */
+	uint32_t first_prio:3;	/* Priority of first VLAN tag in the incoming packet. Valid only when ;cvlan_tag==1 or svlan_tag==1 */
+	uint32_t first_cfi:1;	/* CFI bit of first VLAN tag in the incoming packet. Valid only when ;cvlan_tag==1 or svlan_tag==1 */
+	uint32_t first_vid:12;	/* VLAN ID of first VLAN tag in the incoming packet. Valid only ;when cvlan_tag==1 or svlan_tag==1 */
 	uint32_t ip_protocol:8;	/* IP protocol */
-	uint32_t tcp_dport:16;	/* TCP destination port. ;tcp and udp sport/dport are mutually exclusive */
+	uint32_t ip_dscp:6;	/* Differentiated Services Code Point derived from Traffic Class/;TOS field of IPv6/v4 */
+	uint32_t ip_ecn:2;	/* Explicit Congestion Notification derived from Traffic Class/TOS ;field of IPv6/v4 */
+	uint32_t cvlan_tag:1;	/* The first vlan in the packet is c-vlan (0x8100). cvlan_tag and ;svlan_tag cannot be set together */
+	uint32_t svlan_tag:1;	/* The first vlan in the packet is s-vlan (0x8a88). cvlan_tag and ;svlan_tag cannot be set together */
+	uint32_t frag:1;	/* Packet is an IP fragment */
+	uint32_t ip_version:4;	/* IP version */
+	uint32_t tcp_flags:9;	/* TCP flags. ;Bit 0: FIN;Bit 1: SYN;Bit 2: RST;Bit 3: PSH;Bit 4: ACK;Bit 5: URG;Bit 6: ECE;Bit 7: CWR;Bit 8: NS */
 	uint32_t tcp_sport:16;	/* TCP source port.;tcp and udp sport/dport are mutually exclusive */
+	uint32_t tcp_dport:16;	/* TCP destination port. ;tcp and udp sport/dport are mutually exclusive */
+	uint32_t reserved_at_c0:16;
 	uint32_t ipv4_ihl:4;
 	uint32_t l3_ok:1;
 	uint32_t l4_ok:1;
 	uint32_t ipv4_checksum_ok:1;
 	uint32_t l4_checksum_ok:1;
 	uint32_t ip_ttl_hoplimit:8;
-	uint32_t udp_dport:16;	/* UDP destination port.;tcp and udp sport/dport are mutually exclusive */
 	uint32_t udp_sport:16;	/* UDP source port.;tcp and udp sport/dport are mutually exclusive */
+	uint32_t udp_dport:16;	/* UDP destination port.;tcp and udp sport/dport are mutually exclusive */
 	uint32_t src_ip_127_96;	/* IPv6 source address of incoming packets ;For IPv4 address use bits 31:0 (rest of the bits are reserved);This field should be qualified by an appropriate ;ethertype */
 	uint32_t src_ip_95_64;	/* IPv6 source address of incoming packets ;For IPv4 address use bits 31:0 (rest of the bits are reserved);This field should be qualified by an appropriate ;ethertype */
 	uint32_t src_ip_63_32;	/* IPv6 source address of incoming packets ;For IPv4 address use bits 31:0 (rest of the bits are reserved);This field should be qualified by an appropriate ;ethertype */
@@ -689,52 +735,68 @@ struct dr_match_spec {
 };
 
 struct dr_match_misc {
-	uint32_t source_sqn:24;			/* Source SQN */
-	uint32_t source_vhca_port:4;
-	uint32_t gre_s_present:1;		/* used with GRE, sequence number exist when gre_s_present == 1 */
-	uint32_t gre_k_present:1;		/* used with GRE, key exist when gre_k_present == 1 */
 	uint32_t gre_c_present:1;		/* used with GRE, checksum exist when gre_c_present == 1 */
+	uint32_t bth_a:1;
+	uint32_t gre_k_present:1;		/* used with GRE, key exist when gre_k_present == 1 */
+	uint32_t gre_s_present:1;		/* used with GRE, sequence number exist when gre_s_present == 1 */
+	uint32_t source_vhca_port:4;
+	uint32_t source_sqn:24;			/* Source SQN */
+	uint32_t source_eswitch_owner_vhca_id:16;
 	uint32_t source_port:16;		/* Source port.;0xffff determines wire port */
-	uint32_t inner_second_vid:12;		/* VLAN ID of first VLAN tag the inner header of the incoming packet. ;Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
-	uint32_t inner_second_cfi:1;		/* CFI bit of first VLAN tag in the inner header of the incoming packet. ;Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
-	uint32_t inner_second_prio:3;		/* Priority of second VLAN tag in the inner header of the incoming ;packet. Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
-	uint32_t outer_second_vid:12;		/* VLAN ID of first VLAN tag the outer header of the incoming packet. ;Valid only when outer_second_cvlan_tag ==1 or outer_sec;ond_svlan_tag ==1 */
-	uint32_t outer_second_cfi:1;		/* CFI bit of first VLAN tag in the outer header of the incoming packet. ;Valid only when outer_second_cvlan_tag ==1 or outer_sec;ond_svlan_tag ==1 */
 	uint32_t outer_second_prio:3;		/* Priority of second VLAN tag in the outer header of the incoming ;packet. Valid only when outer_second_cvlan_tag ==1 or outer_sec;ond_svlan_tag ==1 */
-	uint32_t gre_protocol:16;		/* GRE Protocol (outer) */
-	uint32_t inner_second_svlan_tag:1;	/* The second vlan in the inner header of the packet is s-vlan (0x8a88). ;inner_second_cvlan_tag and inner_second_svlan_tag cannot be set ;together */
-	uint32_t outer_second_svlan_tag:1;	/* The second vlan in the outer header of the packet is s-vlan (0x8a88). ;outer_second_cvlan_tag and outer_second_svlan_tag cannot be set ;together */
-	uint32_t inner_second_cvlan_tag:1;	/* The second vlan in the inner header of the packet is c-vlan (0x8100). ;inner_second_cvlan_tag and inner_second_svlan_tag cannot be set ;together */
+	uint32_t outer_second_cfi:1;		/* CFI bit of first VLAN tag in the outer header of the incoming packet. ;Valid only when outer_second_cvlan_tag ==1 or outer_sec;ond_svlan_tag ==1 */
+	uint32_t outer_second_vid:12;		/* VLAN ID of first VLAN tag the outer header of the incoming packet. ;Valid only when outer_second_cvlan_tag ==1 or outer_sec;ond_svlan_tag ==1 */
+	uint32_t inner_second_prio:3;		/* Priority of second VLAN tag in the inner header of the incoming ;packet. Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
+	uint32_t inner_second_cfi:1;		/* CFI bit of first VLAN tag in the inner header of the incoming packet. ;Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
+	uint32_t inner_second_vid:12;		/* VLAN ID of first VLAN tag the inner header of the incoming packet. ;Valid only when inner_second_cvlan_tag ==1 or inner_sec;ond_svlan_tag ==1 */
 	uint32_t outer_second_cvlan_tag:1;	/* The second vlan in the outer header of the packet is c-vlan (0x8100). ;outer_second_cvlan_tag and outer_second_svlan_tag cannot be set ;together */
-	uint32_t gre_key_l:8;			/* GRE Key [7:0] (outer) */
+	uint32_t inner_second_cvlan_tag:1;	/* The second vlan in the inner header of the packet is c-vlan (0x8100). ;inner_second_cvlan_tag and inner_second_svlan_tag cannot be set ;together */
+	uint32_t outer_second_svlan_tag:1;	/* The second vlan in the outer header of the packet is s-vlan (0x8a88). ;outer_second_cvlan_tag and outer_second_svlan_tag cannot be set ;together */
+	uint32_t inner_second_svlan_tag:1;	/* The second vlan in the inner header of the packet is s-vlan (0x8a88). ;inner_second_cvlan_tag and inner_second_svlan_tag cannot be set ;together */
+	uint32_t outer_emd_tag:1;
+	uint32_t reserved_at_65:11;
+	uint32_t gre_protocol:16;		/* GRE Protocol (outer) */
 	uint32_t gre_key_h:24;			/* GRE Key[31:8] (outer) */
+	uint32_t gre_key_l:8;			/* GRE Key [7:0] (outer) */
 	uint32_t vxlan_vni:24;			/* VXLAN VNI (outer) */
-	uint32_t geneve_oam:1;			/* GENEVE OAM field (outer) */
+	uint32_t bth_opcode:8;			/* Opcode field from BTH header */
 	uint32_t geneve_vni:24;			/* GENEVE VNI field (outer) */
+	uint32_t reserved_at_e4:6;
+	uint32_t geneve_tlv_option_0_exist:1;
+	uint32_t geneve_oam:1;			/* GENEVE OAM field (outer) */
+	uint32_t reserved_at_ec:12;
 	uint32_t outer_ipv6_flow_label:20;	/* Flow label of incoming IPv6 packet (outer) */
+	uint32_t reserved_at_100:12;
 	uint32_t inner_ipv6_flow_label:20;	/* Flow label of incoming IPv6 packet (inner) */
-	uint32_t geneve_protocol_type:16;	/* GENEVE protocol type (outer) */
+	uint32_t reserved_at_120:10;
 	uint32_t geneve_opt_len:6;		/* GENEVE OptLen (outer) */
+	uint32_t geneve_protocol_type:16;	/* GENEVE protocol type (outer) */
+	uint32_t reserved_at_140:8;
 	uint32_t bth_dst_qp:24;			/* Destination QP in BTH header */
+	uint32_t inner_esp_spi;
+	uint32_t outer_esp_spi;
+	uint32_t reserved_at_1a0;
+	uint32_t reserved_at_1c0;
+	uint32_t reserved_at_1e0;
 };
 
 struct dr_match_misc2 {
-	uint32_t outer_first_mpls_ttl:8;		/* First MPLS TTL (outer) */
-	uint32_t outer_first_mpls_s_bos:1;		/* First MPLS S_BOS (outer) */
-	uint32_t outer_first_mpls_exp:3;		/* First MPLS EXP (outer) */
 	uint32_t outer_first_mpls_label:20;		/* First MPLS LABEL (outer) */
-	uint32_t inner_first_mpls_ttl:8;		/* First MPLS TTL (inner) */
-	uint32_t inner_first_mpls_s_bos:1;		/* First MPLS S_BOS (inner) */
-	uint32_t inner_first_mpls_exp:3;		/* First MPLS EXP (inner) */
+	uint32_t outer_first_mpls_exp:3;		/* First MPLS EXP (outer) */
+	uint32_t outer_first_mpls_s_bos:1;		/* First MPLS S_BOS (outer) */
+	uint32_t outer_first_mpls_ttl:8;		/* First MPLS TTL (outer) */
 	uint32_t inner_first_mpls_label:20;		/* First MPLS LABEL (inner) */
-	uint32_t outer_first_mpls_over_gre_ttl:8;	/* last MPLS TTL (outer) */
-	uint32_t outer_first_mpls_over_gre_s_bos:1;	/* last MPLS S_BOS (outer) */
-	uint32_t outer_first_mpls_over_gre_exp:3;	/* last MPLS EXP (outer) */
+	uint32_t inner_first_mpls_exp:3;		/* First MPLS EXP (inner) */
+	uint32_t inner_first_mpls_s_bos:1;		/* First MPLS S_BOS (inner) */
+	uint32_t inner_first_mpls_ttl:8;		/* First MPLS TTL (inner) */
 	uint32_t outer_first_mpls_over_gre_label:20;	/* last MPLS LABEL (outer) */
-	uint32_t outer_first_mpls_over_udp_ttl:8;	/* last MPLS TTL (outer) */
-	uint32_t outer_first_mpls_over_udp_s_bos:1;	/* last MPLS S_BOS (outer) */
-	uint32_t outer_first_mpls_over_udp_exp:3;	/* last MPLS EXP (outer) */
+	uint32_t outer_first_mpls_over_gre_exp:3;	/* last MPLS EXP (outer) */
+	uint32_t outer_first_mpls_over_gre_s_bos:1;	/* last MPLS S_BOS (outer) */
+	uint32_t outer_first_mpls_over_gre_ttl:8;	/* last MPLS TTL (outer) */
 	uint32_t outer_first_mpls_over_udp_label:20;	/* last MPLS LABEL (outer) */
+	uint32_t outer_first_mpls_over_udp_exp:3;	/* last MPLS EXP (outer) */
+	uint32_t outer_first_mpls_over_udp_s_bos:1;	/* last MPLS S_BOS (outer) */
+	uint32_t outer_first_mpls_over_udp_ttl:8;	/* last MPLS TTL (outer) */
 	uint32_t metadata_reg_c_7;			/* metadata_reg_c_7 */
 	uint32_t metadata_reg_c_6;			/* metadata_reg_c_6 */
 	uint32_t metadata_reg_c_5;			/* metadata_reg_c_5 */
@@ -744,7 +806,9 @@ struct dr_match_misc2 {
 	uint32_t metadata_reg_c_1;			/* metadata_reg_c_1 */
 	uint32_t metadata_reg_c_0;			/* metadata_reg_c_0 */
 	uint32_t metadata_reg_a;			/* metadata_reg_a */
-	uint32_t metadata_reg_b;			/* metadata_reg_b */
+	uint32_t reserved_at_1a0;
+	uint32_t reserved_at_1c0;
+	uint32_t reserved_at_1e0;
 };
 
 struct dr_match_misc3 {
@@ -752,22 +816,26 @@ struct dr_match_misc3 {
 	uint32_t outer_tcp_seq_num;
 	uint32_t inner_tcp_ack_num;
 	uint32_t outer_tcp_ack_num;
+	uint32_t reserved_at_80:8;
 	uint32_t outer_vxlan_gpe_vni:24;
-	uint32_t outer_vxlan_gpe_flags:8;
 	uint32_t outer_vxlan_gpe_next_protocol:8;
+	uint32_t outer_vxlan_gpe_flags:8;
+	uint32_t reserved_at_b0:16;
 	uint32_t icmpv4_header_data;
 	uint32_t icmpv6_header_data;
-	uint8_t icmpv6_code;
-	uint8_t icmpv6_type;
-	uint8_t icmpv4_code;
 	uint8_t icmpv4_type;
+	uint8_t icmpv4_code;
+	uint8_t icmpv6_type;
+	uint8_t icmpv6_code;
 	uint32_t geneve_tlv_option_0_data;
 	uint32_t gtpu_teid;
 	uint32_t gtpu_msg_type:8;
 	uint32_t gtpu_msg_flags:8;
+	uint32_t reserved_at_170:16;
 	uint32_t gtpu_dw_2;
 	uint32_t gtpu_first_ext_dw_0;
 	uint32_t gtpu_dw_0;
+	uint32_t reserved_at_1e0;
 };
 
 struct dr_match_misc4 {
@@ -798,7 +866,14 @@ struct dr_match_misc5 {
 	uint32_t tunnel_header_1;
 	uint32_t tunnel_header_2;
 	uint32_t tunnel_header_3;
-	uint32_t reserved[0x8];
+	uint32_t reserved_at_100;
+	uint32_t reserved_at_120;
+	uint32_t reserved_at_140;
+	uint32_t reserved_at_160;
+	uint32_t reserved_at_180;
+	uint32_t reserved_at_1a0;
+	uint32_t reserved_at_1c0;
+	uint32_t reserved_at_1e0;
 };
 
 struct dr_match_param {
@@ -873,6 +948,11 @@ struct dr_devx_caps {
 	uint32_t			log_icm_size;
 	uint8_t				log_modify_hdr_icm_size;
 	uint64_t			hdr_modify_icm_addr;
+	uint32_t			log_modify_pattern_icm_size;
+	uint64_t			hdr_modify_pattern_icm_addr;
+	uint64_t			indirect_encap_icm_base;
+	uint32_t			log_sw_encap_icm_size;
+	uint16_t			max_encap_size;
 	uint32_t			flex_protocols;
 	uint8_t				flex_parser_header_modify;
 	uint8_t				flex_parser_id_icmp_dw0;
@@ -886,6 +966,7 @@ struct dr_devx_caps {
 	uint8_t				flex_parser_id_gtpu_teid;
 	uint8_t				flex_parser_id_gtpu_dw_2;
 	uint8_t				flex_parser_id_gtpu_first_ext_dw_0;
+	uint8_t				flex_parser_ok_bits_supp;
 	uint8_t				definer_supp_checksum;
 	uint8_t				max_ft_level;
 	uint8_t				sw_format_ver;
@@ -899,9 +980,13 @@ struct dr_devx_caps {
 	bool				fdb_sw_owner_v2;
 	struct dr_devx_roce_cap		roce_caps;
 	uint64_t			definer_format_sup;
+	uint16_t			log_header_modify_argument_granularity;
+	uint16_t			log_header_modify_argument_max_alloc;
+	bool				support_modify_argument;
 	bool				prio_tag_required;
 	bool				is_ecpf;
 	struct dr_devx_vports		vports;
+	bool				support_full_tnl_hdr;
 };
 
 struct dr_devx_flow_table_attr {
@@ -974,10 +1059,12 @@ struct dr_domain_rx_tx {
 
 struct dr_domain_info {
 	bool			supp_sw_steering;
-	uint32_t		max_inline_size;
 	uint32_t		max_log_sw_icm_sz;
 	uint32_t		max_log_action_icm_sz;
-	uint32_t		max_send_size;
+	uint32_t		max_log_modify_hdr_pattern_icm_sz;
+	uint32_t                max_log_sw_icm_rehash_sz;
+	uint32_t		max_log_sw_encap_icm_sz;
+	uint32_t                max_send_size;
 	struct dr_domain_rx_tx	rx;
 	struct dr_domain_rx_tx	tx;
 	struct ibv_device_attr_ex attr;
@@ -994,17 +1081,23 @@ struct mlx5dv_dr_domain {
 	struct ibv_context		*ctx;
 	struct dr_ste_ctx		*ste_ctx;
 	struct ibv_pd			*pd;
+	int				pd_num;
 	struct mlx5dv_devx_uar		*uar;
 	enum mlx5dv_dr_domain_type	type;
 	atomic_int			refcount;
 	struct dr_icm_pool		*ste_icm_pool;
 	struct dr_icm_pool		*action_icm_pool;
+	struct dr_ptrn_mngr		*modify_header_ptrn_mngr;
+	struct dr_arg_mngr		*modify_header_arg_mngr;
+	struct dr_icm_pool		*encap_icm_pool;
 	struct dr_send_ring		*send_ring[DR_MAX_SEND_RINGS];
 	struct dr_domain_info		info;
 	struct list_head		tbl_list;
 	uint32_t			flags;
 	/* protect debug lists of all tracked objects */
 	pthread_spinlock_t		debug_lock;
+	/* statistcs */
+	uint32_t num_buddies[DR_ICM_TYPE_MAX];
 };
 
 static inline int dr_domain_nic_lock_init(struct dr_domain_rx_tx *nic_dmn)
@@ -1134,6 +1227,33 @@ struct dr_flow_sampler_restore_tbl {
 	uint16_t			num_of_actions;
 };
 
+struct dr_rewrite_param {
+	struct dr_icm_chunk *chunk;
+	uint8_t *data;
+	uint32_t data_size;
+	uint16_t num_of_actions;
+	uint32_t index;
+};
+
+enum dr_ptrn_type {
+	DR_PTRN_TYP_MODIFY_HDR = DR_ACTION_TYP_MODIFY_HDR,
+	DR_PTRN_TYP_TNL_L3_TO_L2 = DR_ACTION_TYP_TNL_L3_TO_L2,
+};
+
+struct dr_ptrn_obj {
+	struct dr_rewrite_param rewrite_param;
+	atomic_int refcount;
+	struct list_node list;
+	enum dr_ptrn_type type;
+};
+
+struct dr_arg_obj {
+	struct mlx5dv_devx_obj *obj;
+	uint32_t obj_offset;
+	struct list_node list_node;
+	uint32_t log_chunk_size;
+};
+
 struct mlx5dv_dr_action {
 	enum dr_action_type		action_type;
 	atomic_int			refcount;
@@ -1141,16 +1261,18 @@ struct mlx5dv_dr_action {
 		struct {
 			struct mlx5dv_dr_domain	*dmn;
 			bool			is_root_level;
+			uint32_t		args_send_qp;
 			union {
 				struct ibv_flow_action	*flow_action; /* root*/
 				struct {
-					struct dr_icm_chunk	*chunk;
-					uint8_t			*data;
-					uint32_t		data_size;
-					uint16_t		num_of_actions;
-					uint32_t		index;
-					bool			allow_rx;
-					bool			allow_tx;
+					struct dr_rewrite_param	param;
+					uint8_t			single_action_opt:1;
+					uint8_t			allow_rx:1;
+					uint8_t			allow_tx:1;
+					struct {
+						struct dr_ptrn_obj *ptrn;
+						struct dr_arg_obj *arg;
+					} ptrn_arg;
 				};
 			};
 		} rewrite;
@@ -1161,6 +1283,9 @@ struct mlx5dv_dr_action {
 				struct ibv_flow_action	*flow_action; /* root*/
 				struct {
 					struct mlx5dv_devx_obj	*dvo;
+					uint8_t			*data;
+					uint32_t		index;
+					struct dr_icm_chunk	*chunk;
 					uint32_t		reformat_size;
 				};
 			};
@@ -1204,6 +1329,13 @@ struct mlx5dv_dr_action {
 				struct ibv_qp           *qp;
 			};
 		} dest_qp;
+		struct {
+			struct mlx5dv_dr_table *tbl;
+			struct dr_devx_tbl *devx_tbl;
+			struct mlx5dv_steering_anchor *sa;
+			uint64_t rx_icm_addr;
+			uint64_t tx_icm_addr;
+		} root_tbl;
 		struct dr_action_aso	aso;
 		struct mlx5dv_devx_obj	*devx_obj;
 		uint32_t		flow_tag;
@@ -1294,14 +1426,9 @@ int dr_rule_send_update_list(struct list_head *send_ste_list,
 struct dr_icm_chunk {
 	struct dr_icm_buddy_mem *buddy_mem;
 	struct list_node	chunk_list;
-	uint32_t		rkey;
 	uint32_t		num_of_entries;
 	uint32_t		byte_size;
-	uint64_t		icm_addr;
-	uint64_t		mr_addr;
-	/* indicates the index of this chunk in the whole memory,
-	 * used for deleting the chunk from the buddy
-	 */
+	/* segment indicates the index of this chunk in its buddy's memory */
 	uint32_t		seg;
 
 	/* Memory optimisation */
@@ -1315,6 +1442,8 @@ dr_icm_pool_dm_type_to_entry_size(enum dr_icm_type icm_type)
 {
 	if (icm_type == DR_ICM_TYPE_STE)
 		return DR_STE_SIZE;
+	else if (icm_type == DR_ICM_TYPE_ENCAP)
+		return DR_SW_ENCAP_ENTRY_SIZE;
 
 	return DR_MODIFY_ACTION_SIZE;
 }
@@ -1465,7 +1594,9 @@ int dr_devx_modify_qp_rtr2rts(struct ibv_context *ctx,
 			      struct dr_devx_qp_rts_attr *attr);
 int dr_devx_query_gid(struct ibv_context *ctx, uint8_t vhca_port_num,
 		      uint16_t index, struct dr_gid_attr *attr);
-
+struct mlx5dv_devx_obj *dr_devx_create_modify_header_arg(struct ibv_context *ctx,
+							 uint16_t log_obj_range,
+							 uint32_t pd);
 static inline bool dr_is_root_table(struct mlx5dv_dr_table *tbl)
 {
 	return tbl->level == 0;
@@ -1484,6 +1615,10 @@ struct dr_icm_pool *dr_icm_pool_create(struct mlx5dv_dr_domain *dmn,
 void dr_icm_pool_destroy(struct dr_icm_pool *pool);
 int dr_icm_pool_sync_pool(struct dr_icm_pool *pool);
 
+uint64_t dr_icm_pool_get_chunk_icm_addr(struct dr_icm_chunk *chunk);
+uint64_t dr_icm_pool_get_chunk_mr_addr(struct dr_icm_chunk *chunk);
+uint32_t dr_icm_pool_get_chunk_rkey(struct dr_icm_chunk *chunk);
+
 struct dr_icm_chunk *dr_icm_alloc_chunk(struct dr_icm_pool *pool,
 					enum dr_icm_chunk_size chunk_size);
 void dr_icm_free_chunk(struct dr_icm_chunk *chunk);
@@ -1503,7 +1638,7 @@ void dr_ste_set_formated_ste(struct dr_ste_ctx *ste_ctx,
 			     struct dr_htbl_connect_info *connect_info);
 void dr_ste_copy_param(uint8_t match_criteria,
 		       struct dr_match_param *set_param,
-		       struct mlx5dv_flow_match_parameters *mask);
+		       uint64_t *mask_buf, size_t mask_sz, bool clear);
 
 void dr_crc32_init_table(void);
 uint32_t dr_crc32_slice8_calc(const void *input_data, size_t length);
@@ -1585,13 +1720,21 @@ int dr_send_postsend_formated_htbl(struct mlx5dv_dr_domain *dmn,
 				   uint8_t send_ring_idx);
 int dr_send_postsend_action(struct mlx5dv_dr_domain *dmn,
 			    struct mlx5dv_dr_action *action);
+int dr_send_postsend_pattern(struct mlx5dv_dr_domain *dmn,
+			     struct dr_icm_chunk *chunk,
+			     uint16_t num_of_actions,
+			     uint8_t *data);
+int dr_send_postsend_args(struct mlx5dv_dr_domain *dmn, uint64_t arg_id,
+			  uint16_t num_of_actions, uint8_t *actions_data,
+			  uint32_t ring_index);
+
 /* buddy functions & structure */
 struct dr_icm_mr;
 
 struct dr_icm_buddy_mem {
-	bitmap			**bits;
+	unsigned long		**bits;
 	unsigned int		*num_free;
-	bitmap			**set_bit;
+	unsigned long		**set_bit;
 	uint32_t		max_order;
 	struct list_node	list_node;
 	struct dr_icm_mr	*icm_mr;
@@ -1606,15 +1749,48 @@ struct dr_icm_buddy_mem {
 	 * sync_ste command sets them free.
 	 */
 	struct list_head	hot_list;
+
+	/* Memory optimization */
+	struct dr_ste		*ste_arr;
+	struct list_head	*miss_list;
+	uint8_t			*hw_ste_arr;
 	/* HW STE cache entry size */
 	uint8_t                 hw_ste_sz;
 };
 
+bool dr_domain_is_support_modify_hdr_cache(struct mlx5dv_dr_domain *dmn);
+struct dr_ptrn_mngr *
+dr_ptrn_mngr_create(struct mlx5dv_dr_domain *dmn);
+void dr_ptrn_mngr_destroy(struct dr_ptrn_mngr *mngr);
+struct dr_ptrn_obj *
+dr_ptrn_cache_get_pattern(struct dr_ptrn_mngr *mngr,
+			  enum dr_ptrn_type type,
+			  uint16_t num_of_actions,
+			  uint8_t *data);
+void dr_ptrn_cache_put_pattern(struct dr_ptrn_mngr *mngr,
+			       struct dr_ptrn_obj *pattern);
+int dr_ptrn_sync_pool(struct dr_ptrn_mngr *ptrn_mngr);
+
+struct dr_arg_mngr*
+dr_arg_mngr_create(struct mlx5dv_dr_domain *dmn);
+void dr_arg_mngr_destroy(struct dr_arg_mngr *mngr);
+struct dr_arg_obj *dr_arg_get_obj(struct dr_arg_mngr *mngr,
+				  uint16_t num_of_actions,
+				  uint8_t *data);
+void dr_arg_put_obj(struct dr_arg_mngr *mngr, struct dr_arg_obj *arg_obj);
+uint32_t dr_arg_get_object_id(struct dr_arg_obj *arg_obj);
+bool dr_domain_is_support_sw_encap(struct mlx5dv_dr_domain *dmn);
+
 int dr_buddy_init(struct dr_icm_buddy_mem *buddy, uint32_t max_order);
 void dr_buddy_cleanup(struct dr_icm_buddy_mem *buddy);
 int dr_buddy_alloc_mem(struct dr_icm_buddy_mem *buddy, int order);
 void dr_buddy_free_mem(struct dr_icm_buddy_mem *buddy, uint32_t seg, int order);
 
+void dr_ste_free_modify_hdr(struct mlx5dv_dr_action *action);
+int dr_ste_alloc_modify_hdr(struct mlx5dv_dr_action *action);
+int dr_ste_alloc_encap(struct mlx5dv_dr_action *action);
+void dr_ste_free_encap(struct mlx5dv_dr_action *action);
+
 void dr_vports_table_add_wire(struct dr_devx_vports *vports);
 void dr_vports_table_del_wire(struct dr_devx_vports *vports);
 struct dr_devx_vport_cap *dr_vports_table_get_vport_cap(struct dr_devx_caps *caps,
