@@ -69,6 +69,7 @@ DCI_TEST_BAD_FLOW_WITHOUT_RESET = 2
 IB_SMP_ATTR_PORT_INFO = 0x0015
 IB_MGMT_CLASS_SUBN_LID_ROUTED = 0x01
 IB_MGMT_METHOD_GET = 0x01
+DB_BF_DBR_LESS_BUF_OFFSET = 0x600
 
 
 class PortStatus:
@@ -127,11 +128,6 @@ class Mlx5DcResources(RoCETrafficResources):
             self.qps[i].to_rts(attr)
         self.dct_qp.to_rtr(attr)
 
-    def pre_run(self, rpsns, rqps_num):
-        self.rpsns = rpsns
-        self.rqps_num = rqps_num
-        self.to_rts()
-
     def create_context(self):
         mlx5dv_attr = Mlx5DVContextAttr()
         try:
@@ -142,7 +138,8 @@ class Mlx5DcResources(RoCETrafficResources):
             raise unittest.SkipTest('Opening mlx5 context is not supported')
 
     def create_mr(self):
-        access = e.IBV_ACCESS_REMOTE_WRITE | e.IBV_ACCESS_LOCAL_WRITE
+        access = e.IBV_ACCESS_REMOTE_WRITE | e.IBV_ACCESS_LOCAL_WRITE | \
+                 e.IBV_ACCESS_REMOTE_ATOMIC | e.IBV_ACCESS_REMOTE_READ
         self.mr = MR(self.pd, self.msg_size, access)
 
     def create_qp_cap(self):
@@ -151,7 +148,8 @@ class Mlx5DcResources(RoCETrafficResources):
     def create_qp_attr(self):
         qp_attr = QPAttr(port_num=self.ib_port)
         set_rnr_attributes(qp_attr)
-        qp_access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE
+        qp_access = e.IBV_ACCESS_LOCAL_WRITE | e.IBV_ACCESS_REMOTE_WRITE | \
+                    e.IBV_ACCESS_REMOTE_ATOMIC | e.IBV_ACCESS_REMOTE_READ
         qp_attr.qp_access_flags = qp_access
         gr = GlobalRoute(dgid=self.ctx.query_gid(self.ib_port, self.gid_index),
                          sgid_index=self.gid_index)
@@ -397,7 +395,7 @@ class Mlx5DcStreamsRes(Mlx5DcResources):
         :return: None
         """
         import tests.utils as u
-        send_op = e.IBV_QP_EX_WITH_SEND
+        send_op = e.IBV_WR_SEND
         ah_client = u.get_global_ah(client, gid_idx, port)
         s_recv_wr = u.get_recv_wr(server)
         c_recv_wr = u.get_recv_wr(client)
@@ -417,8 +415,10 @@ class Mlx5DcStreamsRes(Mlx5DcResources):
                     if client.bad_flow_handling(qp_idx, e.IBV_WC_SUCCESS, True):
                         continue
                     raise ex
-                if client.bad_flow_handling(qp_idx, wcs[0].status, True):
-                    continue
+                else:
+                    if wcs[0].status != e.IBV_WC_SUCCESS and \
+                            client.bad_flow_handling(qp_idx, wcs[0].status, True):
+                        continue
 
                 u.poll_cq(server.cq)
                 u.post_recv(server, s_recv_wr, qp_idx=qp_idx)
@@ -480,9 +480,12 @@ class Mlx5DevxRcResources(BaseResources):
     The class currently supports post send with immediate, but can be
     easily extended to support other opcodes in the future.
     """
-    def __init__(self, dev_name, ib_port, gid_index, msg_size=1024, activate_port_state=False):
+    def __init__(self, dev_name, ib_port, gid_index, msg_size=1024, activate_port_state=False,
+                 send_dbr_mode=0):
+        from tests.mlx5_prm_structs import SendDbrMode
         super().__init__(dev_name, ib_port, gid_index)
         self.umems = {}
+        self.send_dbr_mode = send_dbr_mode
         self.msg_size = msg_size
         self.num_msgs = 1000
         self.imm = 0x03020100
@@ -525,8 +528,13 @@ class Mlx5DevxRcResources(BaseResources):
                     raise PyverbsRDMAError('Could not change the port state to UP')
                 time.sleep(1)
                 mad_port_state = self.query_port_state_with_mads(ib_port)
+        if self.send_dbr_mode != SendDbrMode.DBR_VALID:
+            self.check_cap_send_dbr_mode()
         self.init_resources()
 
+    def get_wqe_data_segment(self):
+        return WqeDataSeg(self.mr.length, self.mr.lkey, self.mr.buf)
+
     def change_port_state_with_registers(self, state):
         from tests.mlx5_prm_structs import PaosReg
         paos_in = PaosReg(local_port=self.ib_port, admin_status=state, ase=1)
@@ -567,6 +575,36 @@ class Mlx5DevxRcResources(BaseResources):
             raise PyverbsRDMAError(f'Failed to send MAD with syndrome ({mad_ifc_out.syndrome})')
         return mad_ifc_out.mad
 
+    def check_cap_send_dbr_mode(self):
+        """
+        Check the capability of the dbr less.
+        If the HCA cap have HCA cap 2, check if in HCA cap2 0x20(HCA CAP 2) + 0x1(current)
+        have the send_dbr_mode_no_dbr_ext.
+        """
+        from tests.mlx5_prm_structs import QueryCmdHcaCap2Out, \
+            QueryHcaCapIn, QueryCmdHcaCapOut, QueryHcaCapOp, QueryHcaCapMod, SendDbrMode
+        self.create_context()
+        query_cap_in = QueryHcaCapIn(op_mod=0x1)
+        query_cap_out = QueryCmdHcaCapOut(self.ctx.devx_general_cmd(
+            query_cap_in, len(QueryCmdHcaCapOut())))
+        if query_cap_out.status:
+            raise PyverbsRDMAError('Failed to query general HCA CAPs with syndrome '
+                                   f'({query_cap_out.syndrome}')
+
+        if not query_cap_out.capability.hca_cap_2:
+            raise unittest.SkipTest("The device doesn't support general HCA CAPs 2")
+        query_cap2_in = QueryHcaCapIn(op_mod=(QueryHcaCapOp.HCA_CAP_2 << 0x1) | \
+                                              QueryHcaCapMod.CURRENT)
+        query_cap2_out = QueryCmdHcaCap2Out(self.ctx.devx_general_cmd(
+            query_cap2_in, len(QueryCmdHcaCap2Out())))
+        if self.send_dbr_mode == SendDbrMode.NO_DBR_EXT and \
+                not query_cap2_out.capability.send_dbr_mode_no_dbr_ext:
+            raise unittest.SkipTest("The device doesn't support send_dbr_mode_no_dbr_ext cap")
+
+        if self.send_dbr_mode == SendDbrMode.NO_DBR_INT and \
+                not query_cap2_out.capability.send_dbr_mode_no_dbr_int:
+            raise unittest.SkipTest("The device doesn't support send_dbr_mode_no_dbr_int cap")
+
     def init_resources(self):
         if not self.is_eth():
             self.query_lid()
@@ -715,7 +753,7 @@ class Mlx5DevxRcResources(BaseResources):
                     log_rq_size=log_rq_size, log_sq_size=log_sq_size, ts_format=0x1,
                     log_rq_stride=log_rq_stride, uar_page=self.uar['qp'].page_id,
                     cqn_snd=cqn, cqn_rcv=cqn, dbr_umem_id=self.umems['qp_dbr'].umem_id,
-                    dbr_umem_valid=1)
+                    dbr_umem_valid=1, send_dbr_mode=self.send_dbr_mode)
         cmd_in = CreateQpIn(sw_qpc=qpc, wq_umem_id=self.umems['qp'].umem_id,
                             wq_umem_valid=1)
         self.qp = Mlx5DevxObj(self.ctx, cmd_in, len(CreateQpOut()))
@@ -795,26 +833,33 @@ class Mlx5DevxRcResources(BaseResources):
         building the control/data segments, updating and ringing the dbr,
         updating the producer indexes, etc.
         """
+        from tests.mlx5_prm_structs import SendDbrMode
+        buffer_address = self.uar['qp'].reg_addr
+        if self.send_dbr_mode == SendDbrMode.NO_DBR_EXT:
+            # Address of DB blueflame register
+            buffer_address = self.uar['qp'].base_addr + DB_BF_DBR_LESS_BUF_OFFSET
+
         idx = self.qattr.sq.post_idx if self.qattr.sq.post_idx < self.qattr.sq.wqe_num else 0
         buf_offset = self.qattr.sq.offset + (idx << dve.MLX5_SEND_WQE_SHIFT)
         # Prepare WQE
         imm_be32 = struct.unpack("<I", struct.pack(">I", self.imm + self.qattr.sq.post_idx))[0]
         ctrl_seg = WqeCtrlSeg(imm=imm_be32, fm_ce_se=dve.MLX5_WQE_CTRL_CQ_UPDATE)
-        data_seg = WqeDataSeg(self.mr.length, self.mr.lkey, self.mr.buf)
+        data_seg = self.get_wqe_data_segment()
         ctrl_seg.opmod_idx_opcode = (self.qattr.sq.post_idx & 0xffff) << 8 | dve.MLX5_OPCODE_SEND_IMM
-        size_in_octowords = int((ctrl_seg.sizeof() +  data_seg.sizeof()) / 16)
+        size_in_octowords = int((ctrl_seg.sizeof() + data_seg.sizeof()) / 16)
         ctrl_seg.qpn_ds = self.qpn << 8 | size_in_octowords
         Wqe([ctrl_seg, data_seg], self.umems['qp'].umem_addr + buf_offset)
         self.qattr.sq.post_idx += int((size_in_octowords * 16 +
                                        dve.MLX5_SEND_WQE_BB - 1) / dve.MLX5_SEND_WQE_BB)
         # Make sure descriptors are written
         dma.udma_to_dev_barrier()
-        # Update the doorbell record
-        mem.writebe32(self.umems['qp_dbr'].umem_addr,
-                      self.qattr.sq.post_idx & 0xffff, dve.MLX5_SND_DBR)
-        dma.udma_to_dev_barrier()
+        if not self.send_dbr_mode:
+            # Update the doorbell record
+            mem.writebe32(self.umems['qp_dbr'].umem_addr,
+                          self.qattr.sq.post_idx & 0xffff, dve.MLX5_SND_DBR)
+            dma.udma_to_dev_barrier()
         # Ring the doorbell and post the WQE
-        dma.mmio_write64_as_be(self.uar['qp'].reg_addr, mem.read64(ctrl_seg.addr))
+        dma.mmio_write64_as_be(buffer_address, mem.read64(ctrl_seg.addr))
 
     def post_recv(self):
         """
@@ -824,7 +869,7 @@ class Mlx5DevxRcResources(BaseResources):
         """
         buf_offset = self.qattr.rq.offset + self.qattr.rq.wqe_size * self.qattr.rq.head
         # Prepare WQE
-        data_seg = WqeDataSeg(self.mr.length, self.mr.lkey, self.mr.buf)
+        data_seg = self.get_wqe_data_segment()
         Wqe([data_seg], self.umems['qp'].umem_addr + buf_offset)
         # Update indexes
         self.qattr.rq.post_idx += 1
@@ -909,7 +954,7 @@ class Mlx5DevxTrafficBase(Mlx5RDMATestCase):
                             self.server.lid, self.mac_addr)
 
     def send_imm_traffic(self):
-        self.client.mr.write('c' * self.client.msg_size, self.client.msg_size)
+        self.client.mem_write('c' * self.client.msg_size, self.client.msg_size)
         for _ in range(self.client.num_msgs):
             cons_idx = self.client.qattr.cq.cons_idx
             self.server.post_recv()
@@ -922,7 +967,7 @@ class Mlx5DevxTrafficBase(Mlx5RDMATestCase):
             recv_cqe = self.server.poll_cq()
             self.assertEqual(recv_cqe.opcode, dve.MLX5_CQE_RESP_SEND_IMM,
                              'Unexpected CQE opcode')
-            msg_received = self.server.mr.read(self.server.msg_size, 0)
+            msg_received = self.server.mem_read()
             # Validate data (of received message and immediate value)
             tests.utils.validate(msg_received, True, self.server.msg_size)
             imm_inval_pkey = recv_cqe.imm_inval_pkey
@@ -930,8 +975,7 @@ class Mlx5DevxTrafficBase(Mlx5RDMATestCase):
                 imm_inval_pkey = int.from_bytes(
                     imm_inval_pkey.to_bytes(4, byteorder='big'), 'little')
             self.assertEqual(imm_inval_pkey, self.client.imm + cons_idx)
-            self.server.mr.write('s' * self.server.msg_size,
-                                 self.server.msg_size)
+            self.server.mem_write('s' * self.server.msg_size, self.server.msg_size)
 
 
 class Mlx5RcResources(RCResources):
