@@ -12,7 +12,9 @@ from tests.base import PyverbsAPITestCase, RDMATestCase, UDResources
 from pyverbs.pyverbs_error import PyverbsRDMAError
 from pyverbs.base import PyverbsRDMAErrno
 from pyverbs.cq import CompChannel, CQ
+import tests.irdma_base as irdma
 from pyverbs.qp import QPCap
+import pyverbs.device as d
 import tests.utils as u
 
 
@@ -40,11 +42,9 @@ class CQAPITest(PyverbsAPITestCase):
     """
     def setUp(self):
         super().setUp()
-        self.ctx, attr, _ = self.devices[0]
-        self.max_cqe = attr.max_cqe
 
     def test_create_cq(self):
-        for cq_size in [1, self.max_cqe/2, self.max_cqe]:
+        for cq_size in [1, self.attr.max_cqe/2, self.attr.max_cqe]:
             for comp_vector in range(0, min(2, self.ctx.num_comp_vectors)):
                 try:
                     cq = CQ(self.ctx, cq_size, None, None, comp_vector)
@@ -59,17 +59,21 @@ class CQAPITest(PyverbsAPITestCase):
 
 
     def test_create_cq_with_comp_channel(self):
-        for cq_size in [1, self.max_cqe/2, self.max_cqe]:
-            cc = CompChannel(self.ctx)
-            CQ(self.ctx, cq_size, None, cc, 0)
-            cc.close()
+        for cq_size in [1, self.attr.max_cqe/2, self.attr.max_cqe]:
+            try:
+                cc = CompChannel(self.ctx)
+                CQ(self.ctx, cq_size, None, cc, 0)
+                cc.close()
+            except PyverbsRDMAError as ex:
+                if ex.error_code == errno.EOPNOTSUPP:
+                    raise unittest.SkipTest(f'CQ with completion channel is not supported')
 
     def test_create_cq_bad_flow(self):
         """
         Test ibv_create_cq() with wrong comp_vector / number of cqes
         """
         with self.assertRaises(PyverbsRDMAError) as ex:
-            CQ(self.ctx, self.max_cqe + 1, None, None, 0)
+            CQ(self.ctx, self.attr.max_cqe + 1, None, None, 0)
         self.assertEqual(ex.exception.error_code, errno.EINVAL)
 
         with self.assertRaises(PyverbsRDMAError) as ex:
@@ -87,15 +91,6 @@ class CQTest(RDMATestCase):
         self.server = None
         self.client = None
 
-    def create_players(self, resource, **resource_arg):
-        self.client = resource(**self.dev_info, **resource_arg)
-        self.server = resource(**self.dev_info, **resource_arg)
-        self.client.pre_run(self.server.psns, self.server.qps_num)
-        self.server.pre_run(self.client.psns, self.client.qps_num)
-        self.traffic_args = {'client': self.client, 'server': self.server,
-                             'iters': self.iters, 'gid_idx': self.gid_index,
-                             'port': self.ib_port}
-
     def test_resize_cq(self):
         """
         Test resize CQ, start with specific value and then increase and decrease
@@ -122,6 +117,7 @@ class CQTest(RDMATestCase):
                         f'The actual CQ size ({self.client.cq.cqe}) is less '
                         'than guaranteed ({new_cq_size})')
 
+        irdma.skip_if_irdma_dev(d.Context(name=self.dev_name))
         # Fill the CQ entries except one for avoid cq_overrun warnings.
         send_wr, _ = u.get_send_elements(self.client, False)
         ah_client = u.get_global_ah(self.client, self.gid_index, self.ib_port)
