@@ -36,44 +36,30 @@ class DCTest(Mlx5RDMATestCase):
         """
         Exchange the remote attributes between the server and the client.
         """
-        self.server.rkey = self.client.mr.rkey
-        self.server.raddr = self.client.mr.buf
-        self.client.rkey = self.server.mr.rkey
-        self.client.raddr = self.server.mr.buf
+        super().sync_remote_attr()
         self.client.remote_dct_num = self.server.dct_qp.qp_num
         self.server.remote_dct_num = self.client.dct_qp.qp_num
 
-    def create_players(self, resource, bad_flow=DCI_TEST_GOOD_FLOW, **resource_arg):
-        """
-        Init DC tests resources.
-        :param resource: The RDMA resources to use.
-        :param bad_flow: Test bad flows (relevant for DCS tests only)
-        :param resource_arg: Dict of args that specify the resource specific
-        attributes.
-        :return: None
-        """
-        self.client = resource(**self.dev_info, **resource_arg)
-        self.server = resource(**self.dev_info, **resource_arg)
-        if bad_flow:
-            self.client.set_bad_flow(bad_flow)
-        self.client.pre_run(self.server.psns, self.server.qps_num)
-        self.server.pre_run(self.client.psns, self.client.qps_num)
-        self.sync_remote_attr()
-        self.traffic_args = {'client': self.client, 'server': self.server,
-                             'iters': self.iters, 'gid_idx': self.gid_index,
-                             'port': self.ib_port}
-
     def test_dc_rdma_write(self):
         self.create_players(Mlx5DcResources, qp_count=2,
                             send_ops_flags=e.IBV_QP_EX_WITH_RDMA_WRITE)
         u.rdma_traffic(**self.traffic_args, new_send=True,
-                       send_op=e.IBV_QP_EX_WITH_RDMA_WRITE)
+                       send_op=e.IBV_WR_RDMA_WRITE)
 
     def test_dc_send(self):
         self.create_players(Mlx5DcResources, qp_count=2,
                             send_ops_flags=e.IBV_QP_EX_WITH_SEND)
         u.traffic(**self.traffic_args, new_send=True,
-                  send_op=e.IBV_QP_EX_WITH_SEND)
+                  send_op=e.IBV_WR_SEND)
+
+    def test_dc_atomic(self):
+        self.create_players(Mlx5DcResources, qp_count=2,
+                            send_ops_flags=e.IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)
+        client_max_log = self.client.ctx.query_mlx5_device().max_dc_rd_atom
+        server_max_log = self.server.ctx.query_mlx5_device().max_dc_rd_atom
+        u.atomic_traffic(**self.traffic_args, new_send=True,
+                         send_op=e.IBV_WR_ATOMIC_FETCH_AND_ADD,
+                         client_wr=client_max_log, server_wr=server_max_log)
 
     def test_dc_ah_to_qp_mapping(self):
         self.create_players(Mlx5DcResources, qp_count=2,
@@ -86,7 +72,7 @@ class DCTest(Mlx5RDMATestCase):
                 raise unittest.SkipTest('Mapping AH to QP is not supported')
             raise ex
         u.traffic(**self.traffic_args, new_send=True,
-                  send_op=e.IBV_QP_EX_WITH_SEND)
+                  send_op=e.IBV_WR_SEND)
 
     def check_odp_dc_support(self):
         """
@@ -103,7 +89,7 @@ class DCTest(Mlx5RDMATestCase):
         self.create_players(OdpDc, qp_count=2, send_ops_flags=send_ops_flag)
         self.check_odp_dc_support()
         u.traffic(**self.traffic_args, new_send=True,
-                  send_op=e.IBV_QP_EX_WITH_SEND)
+                  send_op=e.IBV_WR_SEND)
 
     def test_dc_rdma_write_stream(self):
         """
@@ -117,7 +103,7 @@ class DCTest(Mlx5RDMATestCase):
         self.create_players(Mlx5DcStreamsRes, qp_count=2,
                             send_ops_flags=e.IBV_QP_EX_WITH_RDMA_WRITE)
         u.rdma_traffic(**self.traffic_args, new_send=True,
-                       send_op=e.IBV_QP_EX_WITH_RDMA_WRITE)
+                       send_op=e.IBV_WR_RDMA_WRITE)
 
     def test_dc_send_stream_bad_flow(self):
         """
@@ -126,8 +112,9 @@ class DCTest(Mlx5RDMATestCase):
         In the end, the test verifies that the number of errors is as expected.
         :raises SkipTest: In case DCI is not supported with HW
         """
-        self.create_players(Mlx5DcStreamsRes, bad_flow=DCI_TEST_BAD_FLOW_WITH_RESET,
+        self.create_players(Mlx5DcStreamsRes,
                             qp_count=1, send_ops_flags=e.IBV_QP_EX_WITH_SEND)
+        self.client.set_bad_flow(DCI_TEST_BAD_FLOW_WITH_RESET)
         self.client.traffic_with_bad_flow(**self.traffic_args)
 
     def test_dc_send_stream_bad_flow_qp(self):
@@ -141,6 +128,7 @@ class DCTest(Mlx5RDMATestCase):
         :raises SkipTest: In case DCI is not supported with HW
         """
         self.iters = 20
-        self.create_players(Mlx5DcStreamsRes, bad_flow=DCI_TEST_BAD_FLOW_WITHOUT_RESET,
+        self.create_players(Mlx5DcStreamsRes,
                             qp_count=1, send_ops_flags=e.IBV_QP_EX_WITH_SEND)
+        self.client.set_bad_flow(DCI_TEST_BAD_FLOW_WITHOUT_RESET)
         self.client.traffic_with_bad_flow(**self.traffic_args)
