@@ -21,3 +21,14 @@ class Mlx5DevxRcTrafficTest(Mlx5DevxTrafficBase):
         self.create_players(Mlx5DevxRcResources)
         # Send traffic
         self.send_imm_traffic()
+
+    def test_devx_rc_qp_send_imm_doorbell_less_traffic(self):
+        """
+        Creates two DevX RC QPs with dbr less ext and modifies them to RTS state.
+        Then does SEND_IMM traffic.
+        """
+        from tests.mlx5_prm_structs import SendDbrMode
+
+        self.create_players(Mlx5DevxRcResources, send_dbr_mode=SendDbrMode.NO_DBR_EXT)
+        # Send traffic
+        self.send_imm_traffic()
